import numpy

from tiled.adapters.array import ArrayAdapter
from tiled.authenticators import Mode, UserSessionState
from tiled.structures.core import StructureFamily


class Authenticator:
    "This accepts any password and stashes it in session state as 'token'."
    mode = Mode.password

    async def authenticate(self, username: str, password: str) -> UserSessionState:
        return UserSessionState(username, {"token": password})


# This stands in for a secret token issued by the external service.
SERVICE_ISSUED_TOKEN = "secret"


class MockClient:
    def __init__(self, base_url):
        self.base_url = base_url

    # This API (get_contents, get_metadata, get_data) is just made up and not important.
    # Could be anything.

    async def get_metadata(self, url, token):
        # This assert stands in for the mocked service
        # authenticating a request.
        assert token == SERVICE_ISSUED_TOKEN
        return {"metadata": str(url)}

    async def get_contents(self, url, token):
        # This assert stands in for the mocked service
        # authenticating a request.
        assert token == SERVICE_ISSUED_TOKEN
        return ["a", "b", "c"]

    async def get_data(self, url, token):
        # This assert stands in for the mocked service
        # authenticating a request.
        assert token == SERVICE_ISSUED_TOKEN
        return numpy.ones((3, 3))


class Adapter:
    """
    This is the Adapter that should be given to the Tiled service in config.

    The only API it should implement is `with_session_state`, which returns
    the AuthenticatedAdapter that actually accesses the data.
    """

    def __init__(self, base_url, metadata=None):
        self.client = MockClient(base_url)
        self.metadata = metadata

    def with_session_state(self, state):
        return AuthenticatedAdapter(self.client, state["token"], metadata=self.metadata)


class AuthenticatedAdapter:
    structure_family = StructureFamily.container

    def __init__(self, client, token, segments=None, metadata=None):
        self._client = client
        self._token = token
        self._segments = segments or []
        self.metadata = metadata or {}

    async def lookup_adapter(self, segments):
        # The service URL is probably something based on segments, but not
        # literally this....
        metadata_url = "/".join(self._segments + segments)
        data_url = "/".join(self._segments + segments)
        metadata = await self._client.get_metadata(metadata_url, token=self._token)
        data = await self._client.get_data(data_url, token=self._token)
        # This could alternatively be some file-based adapter with HDF5Adapter
        # or something custom, or another AuthenticatedAdapter...
        return ArrayAdapter.from_array(data, metadata=metadata)

    async def keys_range(self, offset, limit):
        url = ...  # based on self._segments
        return await self._client.get_contents(url, token=self._token)
        return ["a", "b", "c"]

    async def items_range(self, offset, limit):
        # Ideally this would be a batched request to the external service.
        return [
            (key, self.lookup_adapter([key])) for key in self._keys_range(offset, limit)
        ]



"""Create index for fulltext search

Revision ID: 1cd99c02d0c7
Revises: a66028395cab
Create Date: 2024-01-24 15:53:12.348880

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects.postgresql import JSONB

# revision identifiers, used by Alembic.
revision = "1cd99c02d0c7"
down_revision = "a66028395cab"
branch_labels = None
depends_on = None

# Make JSONB available in column
JSONVariant = sa.JSON().with_variant(JSONB(), "postgresql")


def upgrade():
    connection = op.get_bind()
    if connection.engine.dialect.name == "postgresql":
        with op.get_context().autocommit_block():
            # There is no sane way to perform this using op.create_index()
            op.execute(
                """
                CREATE INDEX metadata_tsvector_search
                ON nodes
                USING gin (jsonb_to_tsvector('simple', metadata, '["string"]'))
                """
            )


def downgrade():
    # This _could_ be implemented but we will wait for a need since we are
    # still in alpha releases.
    raise NotImplementedError



"""
Read https://prometheus.io/docs/practices/naming/ for naming
conventions for metrics & labels. We generally prefer naming them
`tiled_<noun>_<verb>_<type_suffix>`.
"""

import os
from functools import lru_cache

from fastapi import APIRouter, Request, Response, Security
from prometheus_client import CONTENT_TYPE_LATEST, Histogram, generate_latest

from .authentication import get_current_principal

router = APIRouter()

REQUEST_DURATION = Histogram(
    "tiled_request_duration_seconds",
    "request duration for all HTTP requests",
    ["method", "code", "endpoint"],
)
RESPONSE_SIZE = Histogram(
    "tiled_response_content_length_bytes",
    "response size in bytes for all HTTP requests",
    ["method", "code", "endpoint"],
    buckets=[
        1,
        10,
        100,
        1000,
        10_000,
        100_000,
        1_000_000,
        10_000_000,
        100_000_000,
        1_000_000_000,
        10_000_000_000,
        float("inf"),
    ],
)
ACL_DURATION = Histogram(
    "tiled_acl_duration_seconds",
    "time spent applying access control for all HTTP requests",
    ["method", "code", "endpoint"],
)
READ_DURATION = Histogram(
    "tiled_read_duration_seconds",
    "time spent reading data for all HTTP requests",
    ["method", "code", "endpoint"],
)
TOKENIZE_DURATION = Histogram(
    "tiled_tokenization_duration_seconds",
    "time spent generating content identifier for all HTTP requests",
    ["method", "code", "endpoint"],
)
PACK_DURATION = Histogram(
    "tiled_pack_duration_seconds",
    "time spent packing (serializing) data for all HTTP requests",
    ["method", "code", "endpoint"],
)
COMPRESSION_DURATION = Histogram(
    "tiled_compress_duration_seconds",
    "time spent applying compression for all HTTP requests",
    ["method", "code", "endpoint", "encoding"],
)
COMPRESSION_RATIO = Histogram(
    "tiled_compress_ratio",
    "compression ratio for all HTTP requests",
    ["method", "code", "endpoint", "encoding"],
    buckets=[1, 2.5, 5, 10, 15, 30, 60, 120, float("inf")],
)

# Initialize labels in advance so that the metrics exist (and can be used in
# dashboards and alerts) even if they have not yet occurred.
for code in ["200", "304", "500"]:
    for endpoint in {
        "about",
        "auth",
        "data",
        "entries",
        "metadata",
        "metrics",
        "search",
    }:
        REQUEST_DURATION.labels(code=code, method="GET", endpoint=endpoint)
        RESPONSE_SIZE.labels(code=code, method="GET", endpoint=endpoint)
        ACL_DURATION.labels(code=code, method="GET", endpoint=endpoint)
        READ_DURATION.labels(code=code, method="GET", endpoint=endpoint)
        TOKENIZE_DURATION.labels(code=code, method="GET", endpoint=endpoint)
        PACK_DURATION.labels(code=code, method="GET", endpoint=endpoint)
        for encoding in ["blosc", "gzip", "lz4", "zstd"]:
            COMPRESSION_DURATION.labels(
                code=code, method="GET", endpoint=endpoint, encoding=encoding
            )
            COMPRESSION_RATIO.labels(
                code=code, method="GET", endpoint=endpoint, encoding=encoding
            )


def capture_request_metrics(request, response):
    method = request.method
    code = response.status_code
    metrics = request.state.metrics
    endpoint = getattr(request.state, "endpoint", "unknown")
    REQUEST_DURATION.labels(method=method, code=code, endpoint=endpoint).observe(
        metrics["app"]["dur"]
    )
    RESPONSE_SIZE.labels(method=method, code=code, endpoint=endpoint).observe(
        int(response.headers.get("content-length", 0))
    )
    if "acl" in metrics:
        ACL_DURATION.labels(method=method, code=code, endpoint=endpoint).observe(
            metrics["acl"]["dur"]
        )
    if "read" in metrics:
        READ_DURATION.labels(method=method, code=code, endpoint=endpoint).observe(
            metrics["read"]["dur"]
        )
    if "tok" in metrics:
        TOKENIZE_DURATION.labels(method=method, code=code, endpoint=endpoint).observe(
            metrics["tok"]["dur"]
        )
    if "pack" in metrics:
        PACK_DURATION.labels(method=method, code=code, endpoint=endpoint).observe(
            metrics["read"]["pack"]
        )
    if "compress" in metrics:
        encoding = response.headers["content-encoding"]
        COMPRESSION_DURATION.labels(
            method=method, code=code, endpoint=endpoint, encoding=encoding
        ).observe(metrics["compress"]["dur"])
        COMPRESSION_RATIO.labels(
            method=method, code=code, endpoint=endpoint, encoding=encoding
        ).observe(metrics["compress"]["ratio"])


@lru_cache()
def prometheus_registry():
    """
    Configure prometheus_client.

    This is run the first time the /metrics endpoint is used.
    """
    if os.environ.get("PROMETHEUS_MULTIPROC_DIR"):
        # The multiprocess configuration makes it compatible with gunicorn.
        # https://github.com/prometheus/client_python/#multiprocess-mode-eg-gunicorn
        from prometheus_client import CollectorRegistry
        from prometheus_client.multiprocess import MultiProcessCollector

        registry = CollectorRegistry()
        MultiProcessCollector(registry)  # This has a side effect.
    else:
        from prometheus_client import REGISTRY as registry

    return registry


@router.get("/metrics")
async def metrics(
    request: Request, principal=Security(get_current_principal, scopes=["metrics"])
):
    """
    Prometheus metrics
    """

    request.state.endpoint = "metrics"
    data = generate_latest(prometheus_registry())
    return Response(data, headers={"Content-Type": CONTENT_TYPE_LATEST})



import os
from pathlib import Path

import httpx

from .utils import SerializableLock, handle_error


class CannotRefreshAuthentication(Exception):
    pass


class TiledAuth(httpx.Auth):
    def __init__(self, refresh_url, csrf_token, token_directory):
        self.refresh_url = refresh_url
        self.csrf_token = csrf_token
        self.token_directory = Path(token_directory)
        self.token_directory.mkdir(exist_ok=True, parents=True)
        self._sync_lock = SerializableLock()
        # self._async_lock = asyncio.Lock()
        self.tokens = {}
        self._check_writable_token_directory()

    def _check_writable_token_directory(self):
        if not os.access(self.token_directory, os.W_OK | os.X_OK):
            raise ValueError(
                f"The token_directory {self.token_directory} is not writable."
            )

    def __getstate__(self):
        return (
            self.refresh_url,
            self.csrf_token,
            self.token_directory,
            self._sync_lock,
        )

    def __setstate__(self, state):
        # Omit the cached tokens (self.tokens) from the pickled bundle because:
        # 1. Sometimes users persist pickled data and handle it insecurely.
        # 2. The un-serialized instance of TiledAuth will need to be able to
        #    read/write from token_directory anyway.
        self.tokens = {}
        (refresh_url, csrf_token, token_directory, sync_lock) = state
        self.refresh_url = refresh_url
        self.csrf_token = csrf_token
        self.token_directory = token_directory
        self._sync_lock = sync_lock
        self._check_writable_token_directory()

    def sync_get_token(self, key, reload_from_disk=False):
        if not reload_from_disk:
            # Use in-memory cached copy.
            try:
                return self.tokens[key]
            except Exception:
                pass
        with self._sync_lock:
            filepath = self.token_directory / key
            try:
                with open(filepath, "r") as file:
                    token = file.read()
                    self.tokens[key] = token
                    return token
            except FileNotFoundError:
                return None

    def sync_set_token(self, key, value):
        with self._sync_lock:
            if not isinstance(value, str):
                raise ValueError("Expected string value, got {value!r}")
            filepath = self.token_directory / key
            filepath.touch(mode=0o600)  # Set permissions.
            with open(filepath, "w") as file:
                file.write(value)
            self.tokens[key] = value  # Update cached value.

    def sync_clear_token(self, key):
        with self._sync_lock:
            self.tokens.pop(key, None)
            filepath = self.token_directory / key
            # filepath.unlink(missing_ok=False)  # Python 3.8+
            try:
                filepath.unlink()
            except FileNotFoundError:
                pass
            self.tokens.pop(key, None)  # Clear cached value.

    def sync_auth_flow(self, request, attempt=0):
        access_token = self.sync_get_token("access_token")
        if access_token is not None:
            request.headers["Authorization"] = f"Bearer {access_token}"
            response = yield request
        if (access_token is None) or (response.status_code == 401):
            # Maybe the token cached in memory is stale.
            maybe_new_access_token = self.sync_get_token(
                "access_token", reload_from_disk=True
            )
            if (attempt < 2) and (maybe_new_access_token != access_token):
                return (yield from self.sync_auth_flow(request, attempt=1 + attempt))
            if access_token is not None:
                # The access token is stale or otherwise invalid. Discard.
                self.sync_clear_token("access_token")
            # Begin refresh flow to get a new access token.
            refresh_token = self.sync_get_token("refresh_token", reload_from_disk=True)
            if refresh_token is None:
                raise CannotRefreshAuthentication(
                    "No refresh token was found in token cache. "
                    "Provide fresh credentials. "
                    "For a given client c, use c.context.authenticate()."
                )
            token_request = build_refresh_request(
                self.refresh_url, refresh_token, self.csrf_token
            )
            token_response = yield token_request
            if token_response.status_code == 401:
                # Refreshing the token failed.
                # Discard the expired (or otherwise invalid) refresh_token.
                self.sync_clear_token("refresh_token")
                raise CannotRefreshAuthentication(
                    "Server rejected attempt to refresh token. "
                    "Provide fresh credentials. "
                    "For a given client c, use c.context.authenticate()."
                )
            handle_error(token_response)
            token_response.read()
            tokens = token_response.json()
            # If we get this far, refreshing authentication worked.
            # Store the new refresh token.
            self.sync_set_token("refresh_token", tokens["refresh_token"])
            self.sync_set_token("access_token", tokens["access_token"])
            request.headers["Authorization"] = f"Bearer {tokens['access_token']}"
            yield request

    async def async_get_token(self, key):
        raise NotImplementedError("Async support is planned but not yet implemented.")

    async def async_set_token(self, key, token):
        raise NotImplementedError("Async support is planned but not yet implemented.")

    async def async_clear_token(self, key):
        raise NotImplementedError("Async support is planned but not yet implemented.")

    async def async_auth_flow(self, request):
        raise NotImplementedError("Async support is planned but not yet implemented.")


def build_refresh_request(refresh_url, refresh_token, csrf_token):
    return httpx.Request(
        "POST",
        refresh_url,
        json={"refresh_token": refresh_token},
        # Submit CSRF token in both header and cookie.
        # https://cheatsheetseries.owasp.org/cheatsheets/Cross-Site_Request_Forgery_Prevention_Cheat_Sheet.html#double-submit-cookie
        headers={"x-csrf": csrf_token},
    )



"""
Example code for importing and exporting xdi using tiled

Source: https://github.com/XraySpectroscopy/XAS-Data-Interchange
"""

import collections
import io
import pathlib
import re

import pandas as pd

from tiled.adapters.dataframe import DataFrameAdapter
from tiled.structures.core import Spec
from tiled.utils import path_from_uri


def read_xdi(data_uri, structure=None, metadata=None, specs=None, access_policy=None):
    "Read XDI-formatted file."
    filepath = path_from_uri(data_uri)
    with open(filepath, "r") as file:
        metadata = {}
        fields = collections.defaultdict(dict)

        # if isinstance(f, pathlib.PosixPath):
        #    line = f.read_text().split('\n')[0]
        # else:
        line = file.readline()
        m = re.match(r"#\s*XDI/(\S*)\s*(\S*)?", line)
        if not m:
            raise ValueError(
                f"not an XDI file, no XDI versioning information in first line\n{line}"
            )

        metadata["xdi_version"] = m[1]
        metadata["extra_version"] = m[2]

        field_end_re = re.compile(r"#\s*/{3,}")
        header_end_re = re.compile(r"#\s*-{3,}")

        has_comments = False

        # read header
        for line in file:
            if line[0] != "#":
                raise ValueError(f"reached invalid line in header\n{line}")
            if re.match(field_end_re, line):
                has_comments = True
                break
            elif re.match(header_end_re, line):
                break

            try:
                key, val = line[1:].strip().split(":", 1)
                val = val.strip()
                namespace, tag = key.split(".")
                # TODO coerce to lower case?
            except ValueError:
                print(f"error processing line\n{line}")
                raise

            fields[namespace][tag] = val

        if has_comments:
            comments = ""
            for line in file:
                if re.match(header_end_re, line):
                    break
                comments += line

            metadata["comments"] = comments

        metadata["fields"] = fields

        line = file.readline()
        if line[0] != "#":
            raise ValueError(f"expected column labels. got\n{line}")
        col_labels = line[1:].split()

        # TODO validate

        df = pd.read_table(file, delim_whitespace=True, names=col_labels)

    return DataFrameAdapter.from_pandas(
        df,
        metadata=metadata,
        specs=(specs or []) + [Spec("xdi", version="1.0")],
    )


def write_xdi(df, metadata):
    output = io.StringIO()

    xdi_version = metadata.get("xdi_version")
    extra_version = metadata.get("extra_version")

    output.write(f"# XDI/{xdi_version} {extra_version}\n")

    fields = metadata["fields"]
    for namespace, namespace_dict in fields.items():
        for tag, value in namespace_dict.items():
            output.write(f"# {namespace}.{tag}: {value}\n")

    # write column labels
    columns = list(df.columns)
    output.write("# ")
    output.write(" ".join(columns))
    output.write("\n")

    # write data
    df.to_csv(output, header=False, index=False, sep=" ")
    return output.getvalue()


data = """# XDI/1.0 GSE/1.0
# Column.1: energy eV
# Column.2: i0
# Column.3: itrans
# Column.4: mutrans
# Element.edge: K
# Element.symbol: Cu
# Scan.edge_energy: 8980.0
# Mono.name: Si 111
# Mono.d_spacing: 3.13553
# Beamline.name: 13ID
# Beamline.collimation: none
# Beamline.focusing: yes
# Beamline.harmonic_rejection: rhodium-coated mirror
# Facility.name: APS
# Facility.energy: 7.00 GeV
# Facility.xray_source: APS Undulator A
# Scan.start_time: 2001-06-26T22:27:31
# Detector.I0: 10cm  N2
# Detector.I1: 10cm  N2
# Sample.name: Cu
# Sample.prep: Cu metal foil
# GSE.EXTRA:  config 1
# ///
# Cu foil Room Temperature
# measured at beamline 13-ID
#----
# energy i0 itrans mutrans
  8779.0  149013.7  550643.089065  -1.3070486
  8789.0  144864.7  531876.119084  -1.3006104
  8799.0  132978.7  489591.10592  -1.3033816
  8809.0  125444.7  463051.104096  -1.3059724
  8819.0  121324.7  449969.103983  -1.3107085
  8829.0  119447.7  444386.117562  -1.3138152
  8839.0  119100.7  440176.091039  -1.3072055
  8849.0  117707.7  440448.106567  -1.3195882
  8859.0  117754.7  442302.10637  -1.3233895
  8869.0  117428.7  441944.116528  -1.3253521
  8879.0  117383.7  442810.120466  -1.327693
  8889.0  117185.7  443658.11566  -1.3312944"""


def main():
    pathlib.Path("data").mkdir()
    with open("data/example.xdi", "w") as f:
        f.write(data)


if __name__ == "__main__":
    main()



from typing import Optional, Tuple, Union

import pydantic

from ..structures.sparse import SparseLayout


class COOStructure(pydantic.BaseModel):
    shape: Tuple[int, ...]  # tuple of ints like (3, 3)
    chunks: Tuple[Tuple[int, ...], ...]  # tuple-of-tuples-of-ints like ((3,), (3,))
    dims: Optional[Tuple[str, ...]] = None  # None or tuple of names like ("x", "y")
    resizable: Union[bool, Tuple[bool, ...]] = False
    layout: SparseLayout = SparseLayout.COO


# This may be extended to a Union of structures if more are added.
SparseStructure = COOStructure



import awkward
import awkward.forms

from ..structures.awkward import AwkwardStructure
from ..structures.core import StructureFamily


class AwkwardAdapter:
    structure_family = StructureFamily.awkward

    def __init__(
        self,
        container,
        structure,
        metadata=None,
        specs=None,
        access_policy=None,
    ):
        self.container = container
        self._metadata = metadata or {}
        self._structure = structure
        self.specs = list(specs or [])
        self.access_policy = access_policy

    @classmethod
    def from_array(cls, array, metadata=None, specs=None, access_policy=None):
        form, length, container = awkward.to_buffers(array)
        structure = AwkwardStructure(length=length, form=form.to_dict())
        return cls(
            container,
            structure,
            metadata=metadata,
            specs=specs,
            access_policy=access_policy,
        )

    def metadata(self):
        return self._metadata

    def read_buffers(self, form_keys=None):
        form = awkward.forms.from_dict(self._structure.form)
        keys = [
            key
            for key in form.expected_from_buffers()
            if (form_keys is None)
            or any(key.startswith(form_key) for form_key in form_keys)
        ]
        buffers = {}
        for key in keys:
            buffers[key] = self.container[key]
        return buffers

    def read(self):
        return dict(self.container)

    def write(self, container):
        for form_key, value in container.items():
            self.container[form_key] = value

    def structure(self):
        return self._structure



import json
import uuid as uuid_module

from sqlalchemy import (
    JSON,
    Boolean,
    Column,
    DateTime,
    Enum,
    ForeignKey,
    Integer,
    LargeBinary,
    Table,
    Unicode,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import relationship
from sqlalchemy.sql import func
from sqlalchemy.types import TypeDecorator

from ..server.schemas import PrincipalType
from .base import Base

# Use JSON with SQLite and JSONB with PostgreSQL.
JSONVariant = JSON().with_variant(JSONB(), "postgresql")


class JSONList(TypeDecorator):
    """Represents an immutable structure as a JSON-encoded list.

    Usage::

        JSONList(255)

    """

    impl = Unicode
    cache_ok = True

    def process_bind_param(self, value, dialect):
        # Make sure we don't get passed some iterable like a dict.
        if not isinstance(value, list):
            raise ValueError("JSONList must be given a literal `list` type.")
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class UUID(TypeDecorator):
    """Represents a UUID in a dialect-agnostic way

    Postgres has built-in support but SQLite does not, so we
    just use a 36-character Unicode column.

    We could use 16-byte LargeBinary, which would be more compact
    but we decided it was worth the cost to make the content easily
    inspectable by external database management and development tools.
    """

    impl = Unicode(36)
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is not None:
            if not isinstance(value, uuid_module.UUID):
                raise ValueError(f"Expected uuid.UUID, got {type(value)}")
            return str(value)

    def process_result_value(self, value, dialect):
        if value is not None:
            return uuid_module.UUID(hex=value)


class Timestamped:
    """
    Mixin for providing timestamps of creation and update time.

    These are not used by application code, but they may be useful for
    forensics.
    """

    __mapper_args__ = {"eager_defaults": True}

    time_created = Column(DateTime(timezone=False), server_default=func.now())
    time_updated = Column(
        DateTime(timezone=False), onupdate=func.now()
    )  # null until first update

    def __repr__(self):
        return (
            f"{type(self).__name__}("
            + ", ".join(
                f"{key}={value!r}"
                for key, value in self.__dict__.items()
                if not key.startswith("_")
            )
            + ")"
        )


principal_role_association_table = Table(
    "principal_role_association",
    Base.metadata,
    Column("principal_id", Integer, ForeignKey("principals.id"), primary_key=True),
    Column("role_id", Integer, ForeignKey("roles.id"), primary_key=True),
)


class Principal(Timestamped, Base):
    __tablename__ = "principals"

    # This id is internal, never exposed to the client.
    id = Column(Integer, primary_key=True, index=True, autoincrement=True)
    # This uuid is public.
    uuid = Column(
        UUID,
        index=True,
        nullable=False,
        default=lambda: uuid_module.uuid4(),
    )
    type = Column(Enum(PrincipalType), nullable=False)
    # In the future we may add other information.

    identities = relationship("Identity", back_populates="principal")
    api_keys = relationship("APIKey", back_populates="principal")
    roles = relationship(
        "Role",
        secondary=principal_role_association_table,
        back_populates="principals",
        lazy="joined",
    )
    sessions = relationship("Session", back_populates="principal")


class Identity(Timestamped, Base):
    __tablename__ = "identities"

    # An (id, provider) pair must be unique.
    id = Column(Unicode(255), primary_key=True, nullable=False)
    provider = Column(Unicode(255), primary_key=True, nullable=False)
    principal_id = Column(Integer, ForeignKey("principals.id"), nullable=False)
    latest_login = Column(DateTime(timezone=False), nullable=True)
    # In the future we may add a notion of "primary" identity.

    principal = relationship("Principal", back_populates="identities")


class Role(Timestamped, Base):
    __tablename__ = "roles"

    id = Column(Integer, primary_key=True, index=True, autoincrement=True)
    name = Column(Unicode(255), index=True, unique=True, nullable=False)
    description = Column(Unicode(1023), nullable=True)
    scopes = Column(JSONList(511), nullable=False)
    principals = relationship(
        "Principal", secondary=principal_role_association_table, back_populates="roles"
    )


class APIKey(Timestamped, Base):
    __tablename__ = "api_keys"

    # Store the first_eight characters of the hex-encoded secret.
    # The key holder can use this to identity the key.
    # We do not store the full secret, only its sha256-hashed value.
    # A primary key on (first_eight, hashed_secret) enables
    # fast lookups.
    first_eight = Column(Unicode(8), primary_key=True, index=True, nullable=False)
    hashed_secret = Column(
        LargeBinary(32), primary_key=True, index=True, nullable=False
    )
    expiration_time = Column(DateTime(timezone=False), nullable=True)
    latest_activity = Column(DateTime(timezone=False), nullable=True)
    note = Column(Unicode(1023), nullable=True)
    principal_id = Column(Integer, ForeignKey("principals.id"), nullable=False)
    scopes = Column(JSONList(511), nullable=False)
    # In the future we could make it possible to disable API keys
    # without deleting them from the database, for forensics and
    # record-keeping.

    principal = relationship("Principal", back_populates="api_keys", lazy="joined")


class Session(Timestamped, Base):
    """
    This related to refresh tokens, which have a session uuid ("sid") claim.

    When the client attempts to use a refresh token, we first check
    here to ensure that the "session", which is associated with a chain
    of refresh tokens that came from a single authentication, are still valid.
    """

    __tablename__ = "sessions"

    # This id is internal, never exposed to the client.
    id = Column(Integer, primary_key=True, index=True, autoincrement=True)
    # This uuid is exposed to the client.
    uuid = Column(UUID, index=True, nullable=False, default=uuid_module.uuid4)
    time_last_refreshed = Column(DateTime(timezone=False), nullable=True)
    refresh_count = Column(Integer, nullable=False, default=0)
    expiration_time = Column(DateTime(timezone=False), nullable=False)
    principal_id = Column(Integer, ForeignKey("principals.id"), nullable=False)
    revoked = Column(Boolean, default=False, nullable=False)
    # State allows for custom  authenticator information to be stored in the session.
    state = Column(JSONVariant, nullable=False)
    principal = relationship("Principal", back_populates="sessions", lazy="joined")


class PendingSession(Base):
    """
    This is used only in Device Code Flow.
    """

    __tablename__ = "pending_sessions"

    hashed_device_code = Column(
        LargeBinary(32), primary_key=True, index=True, nullable=False
    )
    user_code = Column(Unicode(8), index=True, nullable=False)
    expiration_time = Column(DateTime(timezone=False), nullable=False)
    session_id = Column(Integer, ForeignKey("sessions.id"), nullable=True)

    session = relationship("Session", lazy="joined")



import numpy
import pytest
from starlette.testclient import TestClient

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..server.app import build_app


@pytest.fixture
def app():
    metadata = {str(i): {str(j): j for j in range(100)} for i in range(100)}
    tree = MapAdapter(
        {
            "compresses_well": ArrayAdapter.from_array(
                numpy.zeros((1000, 1000)), metadata=metadata
            )
        },
    )
    return build_app(tree, authentication={"single_user_api_key": "secret"})
    # In this module we use a raw TestClient instead of tiled.client to omit
    # tiled.client's default headers and other configuration.


def test_gzip_supported(app):
    with TestClient(app=app) as client:
        client.headers["Authorization"] = "Apikey secret"
        client.headers["Accept-Encoding"] = "gzip"
        metadata_response = client.get("/api/v1/search")
        data_response = client.get(
            "/api/v1/array/full/compresses_well", headers={"Accept": "text/csv"}
        )
    assert metadata_response.status_code == 200
    assert data_response.status_code == 200
    assert "gzip" in metadata_response.headers["Content-Encoding"]
    assert "gzip" in data_response.headers["Content-Encoding"]


def test_zstd_preferred(app):
    with TestClient(app=app) as client:
        client.headers["Authorization"] = "Apikey secret"
        client.headers["Accept-Encoding"] = "zstd"
        metadata_response = client.get("/api/v1/search")
        data_response = client.get(
            "/api/v1/array/full/compresses_well", headers={"Accept": "text/csv"}
        )
    assert metadata_response.status_code == 200
    assert data_response.status_code == 200
    assert "zstd" in metadata_response.headers["Content-Encoding"]
    assert "zstd" in data_response.headers["Content-Encoding"]



"""
Adapted from https://raw.githubusercontent.com/obendidi/httpx-cache/main/httpx_cache/transport.py
in accordance with its BSD-3 license
"""
import logging
import typing as tp
from datetime import datetime, timedelta, timezone
from email.utils import parsedate_to_datetime

import attr
import httpx

logger = logging.getLogger(__name__)

_PERMANENT_REDIRECT_STATUSES = (301, 308)


def parse_headers_date(headers_date: tp.Optional[str]) -> tp.Optional[datetime]:
    """Parse a 'Date' header and return it as an optional datetime object.

    If the 'Date' doe not exist return None
    IF there is an error usirng parsing return None

    Args:
        headers: httpx.Headers

    Returns:
        Optional[datetime]
    """
    if not isinstance(headers_date, str):
        return None

    try:
        return parsedate_to_datetime(headers_date)
    except (ValueError, TypeError) as error:
        logger.error(error)
        return None


def parse_cache_control_headers(
    headers: httpx.Headers,
) -> tp.Dict[str, tp.Optional[int]]:
    """Parse cache-control headers.

    Args:
        headers: An instance of httpx headers.

    Returns:
        parsed cache-control headers as dict.
    """

    cache_control: tp.Dict[str, tp.Optional[int]] = {}
    directives = headers.get_list("cache-control", split_commas=True)
    for directive in directives:
        if "=" in directive:
            name, value = directive.split("=", maxsplit=1)
            if value.isdigit():
                cache_control[name] = int(value)
            else:
                cache_control[name] = None
        else:
            cache_control[directive] = None
    return cache_control


@attr.s
class ByteStreamWrapper(httpx.ByteStream):
    """Wrapper around the stream object of an httpx.Response."""

    stream: httpx.ByteStream = attr.ib(kw_only=True)
    callback: tp.Callable[[bytes], tp.Any] = attr.ib(kw_only=True)
    content: bytearray = attr.ib(factory=bytearray, init=False)

    def close(self) -> None:
        """Close stream."""
        self.stream.close()

    async def aclose(self) -> None:
        """Close async stream."""
        await self.stream.aclose()

    def __iter__(self) -> tp.Iterator[bytes]:
        """Iterate over the stream object and store it's chunks in a content.

        After the stream is completed call the callback with content as argument.
        """
        for chunk in self.stream:
            self.content.extend(chunk)
            yield chunk
        self.callback(bytes(self.content))

    async def __aiter__(self) -> tp.AsyncIterator[bytes]:
        """Iterate over the async stream object and store it's chunks in a content.

        After the stream is completed call the async callback with content as argument.
        """
        async for chunk in self.stream:
            self.content.extend(chunk)
            yield chunk
        await self.callback(bytes(self.content))


class CacheControl:
    """Cache controller for httpx-cache.

    Uses 'cache-contol' header direcrives for using/skipping cache.

    If no cache-control directive is set, the cache is used by default (except if there
    is an expires header in the response.)
    """

    def __init__(
        self,
        *,
        cacheable_methods: tp.Tuple[str, ...] = ("GET",),
        cacheable_status_codes: tp.Tuple[int, ...] = (200, 203, 300, 301, 308),
        always_cache: bool = False,
    ) -> None:
        self.cacheable_methods = cacheable_methods
        self.cacheable_status_codes = cacheable_status_codes
        self.always_cache = always_cache

    def is_request_cacheable(self, request: httpx.Request) -> bool:
        """Checks if an httpx request has the necessary requirement to support caching.

        A request is cacheable if:

            - url is absolute
            - method is defined as cacheable (by default only GET methods are cached)
            - request has no 'no-cache' cache-control header directive
            - request has no 'max-age=0' cache-control header directive

        Args:
            request: httpx.Request

        Returns:
            True if request cacheable else False
        """
        if request.url.is_relative_url:
            # logger.debug(
            #     f"Only absolute urls are supported, got '{request.url}'. "
            #     "Request is not cacheable!"
            # )
            return False
        if request.method not in self.cacheable_methods:
            # logger.debug(
            #     f"Request method '{request.method}' is not supported, only "
            #     f"'{self.cacheable_methods}' are supported. Request is not cacheable!"
            # )
            return False
        cc = parse_cache_control_headers(request.headers)
        if "no-cache" in cc or cc.get("max-age") == 0:
            # logger.debug(
            #     "Request cache-control headers has a 'no-cache' directive. "
            #     "Request is not cacheable!"
            # )
            return False
        return True

    def is_response_fresh(
        self, *, request: httpx.Request, response: httpx.Response
    ) -> bool:
        """Checks wether a cached response is fresh or not.

        Args:
            request: httpx.Request
            response: httpx.Response

        Returns:
            True if request is fresh else False
        """

        # check if response is a permanenet redirect
        if response.status_code in _PERMANENT_REDIRECT_STATUSES:
            # logger.debug(
            #     "Cached response with permanent redirect status "
            #     f"'{response.status_code}' is always fresh."
            # )
            return True

        # check that we do have a response Date header
        response_date = parse_headers_date(response.headers.get("date"))

        # extract cache_control for both request and response
        request_cc = parse_cache_control_headers(request.headers)
        response_cc = parse_cache_control_headers(response.headers)

        # get all values we need for freshness eval
        resp_max_age = response_cc.get("max-age")
        req_min_fresh = request_cc.get("min-fresh")
        req_max_age = request_cc.get("max-age")

        # check max-age in response
        if isinstance(req_max_age, int):
            max_freshness_age = timedelta(seconds=req_max_age)
            # logger.debug(
            #     "Evaluating response freshness from request cache-control "
            #     "'max-age' header directive."
            # )

        elif isinstance(resp_max_age, int):
            max_freshness_age = timedelta(seconds=resp_max_age)
            # logger.debug(
            #     "Evaluating response freshness from response cache-control "
            #     "'max-age' header directive."
            # )
        elif "expires" in response.headers and response_date is None:
            # logger.warning(
            #     "Response is missing a valid 'Date' header, couldn't evaluate "
            #     "response freshness. Response is not fresh!"
            # )
            return False
        elif "expires" in response.headers:
            resp_expires = parse_headers_date(response.headers.get("expires"))
            if resp_expires is None:
                # logger.warning(
                #     "Response has an invalid 'Expires' header, couldn't evaluate "
                #     "response freshness. Response is not fresh!"
                # )
                return False

            max_freshness_age = resp_expires - response_date  # type: ignore
            # logger.debug(
            #     "Evaluating response freshness from response 'expires' header."
            # )

        else:
            # logger.debug(
            #     "Request/Response pair has no cache-control headers. Assuming "
            #     "response is fresh!"
            # )
            return True

        if response_date is None:
            # logger.warning(
            #     "Response is missing a valid 'Date' header, couldn't evaluate "
            #     "response freshness. Response is not fresh!"
            # )
            return False

        # get response age (timedelta)
        now = datetime.now(tz=timezone.utc)
        response_age = now - response_date
        if isinstance(req_min_fresh, int):
            # logger.debug(
            #     f"Adjsting response age ({response_age}) using request cache-control "
            #     "'min-fresh' header directive."
            # )
            response_age += timedelta(seconds=req_min_fresh)

        # logger.debug(f"Response age is: {response_age}")
        # logger.debug(f"Response allowed max-age is: {max_freshness_age}")

        if response_age > max_freshness_age:
            # logger.debug("Response is not fresh!")
            return False

        # logger.debug("Response is fresh.")
        return True

    def is_response_cacheable(
        self, *, request: httpx.Request, response: httpx.Response
    ) -> bool:
        """Check if an httpx response is cacheable.

        A respons is cacheable if:

            - response status_code is cacheable
            - request method is cacheable
            - One of:
                - always_cache is True
            OR:
                - Response has no 'no-store' cache-control header
                - Request has no 'no-store' cache-control header

        Args:
            request: httpx.Request
            response: httpx.Response

        Returns:
            wether response is cacheable or not.
        """
        if request.url.is_relative_url:
            # logger.debug(
            #     f"Only absolute urls are supported, got '{request.url}'. "
            #     "Request is not cacheable!"
            # )
            return False

        if request.method not in self.cacheable_methods:
            # logger.debug(
            #     f"Request method '{request.method}' is not supported, only "
            #     f"'{self.cacheable_methods}' are supported. Request is not cacheable!"
            # )
            return False

        if response.status_code not in self.cacheable_status_codes:
            # logger.debug(
            #     f"Response status_code '{response.status_code}' is not cacheable, only "
            #     f"'{self.cacheable_status_codes}' are cacheable. Response is not "
            #     "cacheable!"
            # )
            return False

        # always cache request, eevent if 'no-store' is set as header
        if self.always_cache:
            # logger.debug("Caching Response because 'always_cache' is set to True.'")
            return True

        # extract cache_control for both request and response
        request_cc = parse_cache_control_headers(request.headers)
        response_cc = parse_cache_control_headers(response.headers)

        if "no-store" in request_cc or "no-store" in response_cc:
            # logger.debug(
            #     "Request/Response cache-control headers has a 'no-store' directive. "
            #     "Response is not cacheable!"
            # )
            return False

        return True

    def needs_revalidation(
        self, *, request: httpx.Request, response: httpx.Response
    ) -> bool:
        return "ETag" in response.headers



"""
This tests tiled's validation registry
"""

import numpy as np
import pandas as pd
import pytest

from ..client import Context, from_context
from ..config import merge
from ..server.app import build_app_from_config
from ..structures.core import StructureFamily
from ..validation_registration import ValidationError
from .utils import fail_with_status_code


def lower_case_dict(d):
    out = {}
    modified = False

    for k, v in d.items():
        if isinstance(k, str) and not k.islower():
            out[k.lower()] = v
            modified = True
        else:
            out[k] = v

    return out, modified


def validate_foo(metadata, structure_family, structure, spec):
    if structure_family != StructureFamily.table:
        raise ValidationError(f"structure family for spec {spec} must be 'table'")

    if list(structure.columns) != ["a", "b"]:
        raise ValidationError(f"structure for spec {spec} must have columns ['a', 'b']")

    metadata, metadata_modified = lower_case_dict(metadata)

    if "foo" not in metadata:
        raise ValidationError("metadata for spec {spec} must contain foo")

    if metadata_modified:
        return metadata


@pytest.fixture(scope="module")
def client(tmpdir_module):
    config = {
        "trees": [
            {
                "tree": "tiled.catalog:in_memory",
                "path": "/",
                "args": {"writable_storage": tmpdir_module},
            },
        ],
        "specs": [
            {"spec": "foo", "validator": f"{__name__}:validate_foo"},
            {"spec": "a"},
        ],
    }
    # Check that specs propagate correctly through merging configs.
    merged_config = merge({"filepath_placeholder": config})
    assert merged_config["specs"]
    with Context.from_app(build_app_from_config(merged_config)) as context:
        yield from_context(context)


def test_validators(client):
    # valid example
    df = pd.DataFrame({"a": np.zeros(10), "b": np.zeros(10)})
    client.write_dataframe(df, metadata={"foo": 1}, specs=["foo"])

    with fail_with_status_code(400):
        # not expected structure family
        a = np.ones((5, 7))
        client.write_array(a, metadata={}, specs=["foo"])

    with fail_with_status_code(400):
        # column names are not expected
        df = pd.DataFrame({"x": np.zeros(10), "y": np.zeros(10)})
        client.write_dataframe(df, metadata={}, specs=["foo"])

    with fail_with_status_code(400):
        # missing expected metadata
        df = pd.DataFrame({"a": np.zeros(10), "b": np.zeros(10)})
        client.write_dataframe(df, metadata={}, specs=["foo"])

    metadata = {"id": 1, "foo": "bar"}
    df = pd.DataFrame({"a": np.zeros(10), "b": np.zeros(10)})
    result = client.write_dataframe(df, metadata=metadata, specs=["foo"])
    assert result.metadata == metadata
    result_df = result.read()
    pd.testing.assert_frame_equal(result_df, df)

    metadata_upper = {"ID": 2, "FOO": "bar"}
    metadata_lower, _ = lower_case_dict(metadata_upper)
    result = client.write_dataframe(df, metadata=metadata_upper, specs=["foo"])
    assert result.metadata == metadata_lower
    result_df = result.read()
    pd.testing.assert_frame_equal(result_df, df)


def test_unknown_spec_strict(tmpdir):
    "Test unknown spec rejected for upload."
    config = {
        "trees": [
            {
                "tree": "tiled.catalog:in_memory",
                "path": "/",
                "args": {"writable_storage": tmpdir},
            },
        ],
        "specs": [
            {"spec": "a"},
        ],
        "reject_undeclared_specs": True,
    }
    # Check that specs propagate correctly through merging configs.
    with Context.from_app(build_app_from_config(config)) as context:
        client = from_context(context)
        a = np.ones((5, 7))
        client.write_array(a, metadata={}, specs=["a"])
        with fail_with_status_code(400):
            # unknown spec 'b' should be rejected
            client.write_array(a, metadata={}, specs=["b"])


def test_unknown_spec_permissive(client):
    "Test unknown spec rejected for upload."
    a = np.ones((5, 7))
    client.write_array(a, metadata={}, specs=["a"])
    # unknown spec 'b' should be accepted
    client.write_array(a, metadata={}, specs=["b"])



"""
This tests the feature that is exercised by the --scalable CLI option.
"""
from pathlib import Path

import pytest

from ..config import construct_build_app_kwargs, parse_configs
from ..server.app import UnscalableConfig, build_app

here = Path(__file__).parent.absolute()


@pytest.mark.parametrize(
    "filename, scalable",
    [
        ("config_with_api_key.yml", True),
        ("config_with_secret_keys.yml", True),
        ("config_public_no_authenticator.yml", True),
        ("config_missing_api_key.yml", False),
        ("config_missing_secret_keys.yml", False),
        ("config_missing_secret_keys_public.yml", False),
    ],
)
def test_scalable_config(filename, scalable):
    config_path = here / "test_configs" / filename
    parsed_config = parse_configs(config_path)
    kwargs = construct_build_app_kwargs(parsed_config, source_filepath=config_path)
    if scalable:
        build_app(scalable=True, **kwargs)
        build_app(scalable=False, **kwargs)
    else:
        with pytest.raises(UnscalableConfig):
            build_app(scalable=True, **kwargs)
        build_app(scalable=False, **kwargs)



"""
importing tiled.client should not import any heavy optional dependencies.
"""
import subprocess
import sys

import pytest


@pytest.mark.parametrize(
    "module",
    [
        "numpy",
        "pandas",
        "xarray",
    ],
)
def test_no_heavy_imports(module):
    code = f"import tiled.client; import sys; assert '{module}' not in sys.modules"
    subprocess.check_call([sys.executable, "-c", code])



import numpy
import pytest

from ..utils import safe_json_dump


@pytest.mark.parametrize(
    "test_params",
    argvalues=[
        {"array": numpy.array([0], dtype="i8")},
        {"array": numpy.array([b"test"], dtype="|S4")},
    ],
)
def test_ndarray(test_params):
    safe_json_dump(test_params["array"])


@pytest.mark.parametrize(
    "test_params",
    argvalues=[
        {"array": numpy.array([0], dtype="i8")},
        {"array": numpy.array([b"test"], dtype="|S4")},
    ],
)
def test_nested_ndarray(test_params):
    example_nested_data = {
        "data": [
            {
                "id": "test",
                "attributes": {"metadata": {"test": test_params["array"]}},
            },
            {},
        ]
    }
    safe_json_dump(example_nested_data)



import threading

import dask
import dask.array
import pandas
import xarray

from ..serialization.table import deserialize_arrow
from ..structures.core import Spec
from ..utils import APACHE_ARROW_FILE_MIME_TYPE
from .container import Container
from .utils import handle_error

LENGTH_LIMIT_FOR_WIDE_TABLE_OPTIMIZATION = 1_000_000


class DaskDatasetClient(Container):
    def _repr_pretty_(self, p, cycle):
        """
        Provide "pretty" display in IPython/Jupyter.

        See https://ipython.readthedocs.io/en/stable/config/integrating.html#rich-display
        """
        p.text(f"<{type(self).__name__} {list(self)}>")

    def _ipython_key_completions_(self):
        """
        Provide method for the key-autocompletions in IPython.

        See http://ipython.readthedocs.io/en/stable/config/integrating.html#tab-completion
        """
        return list(self)

    def _build_arrays(self, variables, optimize_wide_table):
        data_vars = {}
        coords = {}
        # Optimization: Download scalar columns in batch as DataFrame.
        # on first access.
        coords_fetcher = _WideTableFetcher(
            self.context.http_client.get, self.item["links"]["full"]
        )
        data_vars_fetcher = _WideTableFetcher(
            self.context.http_client.get, self.item["links"]["full"]
        )
        array_clients = {}
        array_structures = {}
        first_dims = []
        for name, array_client in self.items():
            if (variables is not None) and (name not in variables):
                continue
            array_clients[name] = array_client
            array_structure = array_client.structure()
            array_structures[name] = array_structure
            if array_structure.shape:
                first_dims.append(array_structure.shape[0])
            else:
                first_dims.append(None)
        if len(set(first_dims)) > 1:
            # ragged, not tabular
            optimize_wide_table = False
        for name, array_client in array_clients.items():
            array_structure = array_structures[name]
            shape = array_structure.shape
            spec_names = set(spec.name for spec in array_client.specs)
            if optimize_wide_table and (
                (not shape)  # empty
                or (
                    (shape[0] < LENGTH_LIMIT_FOR_WIDE_TABLE_OPTIMIZATION)
                    and (len(shape) < 2)
                )
            ):
                if "xarray_coord" in spec_names:
                    coords[name] = (
                        array_client.dims,
                        coords_fetcher.register(name, array_client, array_structure),
                    )
                elif "xarray_data_var" in spec_names:
                    data_vars[name] = (
                        array_client.dims,
                        data_vars_fetcher.register(name, array_client, array_structure),
                    )
                else:
                    raise ValueError(
                        "Child nodes of xarray_dataset should include spec "
                        "'xarray_coord' or 'xarray_data_var'."
                    )
            else:
                if "xarray_coord" in spec_names:
                    coords[name] = (array_client.dims, array_client.read())
                elif "xarray_data_var" in spec_names:
                    data_vars[name] = (array_client.dims, array_client.read())
                else:
                    raise ValueError(
                        "Child nodes of xarray_dataset should include spec "
                        "'xarray_coord' or 'xarray_data_var'."
                    )
        return data_vars, coords

    def read(self, variables=None, *, optimize_wide_table=True):
        data_vars, coords = self._build_arrays(variables, optimize_wide_table)
        return xarray.Dataset(
            data_vars=data_vars, coords=coords, attrs=self.metadata["attrs"]
        )


class DatasetClient(DaskDatasetClient):
    def read(self, variables=None, *, optimize_wide_table=True):
        return (
            super()
            .read(variables=variables, optimize_wide_table=optimize_wide_table)
            .load()
        )


URL_CHARACTER_LIMIT = 2000  # number of characters
_EXTRA_CHARS_PER_ITEM = len("&field=")


class _WideTableFetcher:
    def __init__(self, get, link):
        self.get = get
        self.link = link
        self.variables = []
        self._dataframe = None
        # This lock ensures that multiple threads (e.g. dask worker threads)
        # do not prompts us to re-request the same data. Only the first worker
        # to ask for the data should trigger a request.
        self._lock = threading.Lock()

    def register(self, name, array_client, array_structure):
        if self._dataframe is not None:
            raise RuntimeError("Cannot add variables; already fetched.")
        self.variables.append(name)
        # TODO Can we avoid .values here?
        return dask.array.from_delayed(
            dask.delayed(self.dataframe)()[name].values,
            shape=array_structure.shape,
            dtype=array_structure.data_type.to_numpy_dtype(),
        )

    def dataframe(self):
        with self._lock:
            if self._dataframe is None:
                # If self.variables contains many and/or lengthy names,
                # we can bump into the URI size limit commonly imposed by
                # HTTP stacks (e.g. nginx). The HTTP spec does not define a limit,
                # but a common setting is 4K or 8K (for all the headers together).
                # As another reference point, Internet Explorer imposes a
                # 2048-character limit on URLs.
                variables = []
                dataframes = []
                budget = URL_CHARACTER_LIMIT
                budget -= len(self.link)
                # Fetch the variables in batches.
                for variable in self.variables:
                    budget -= _EXTRA_CHARS_PER_ITEM + len(variable)
                    if budget < 0:
                        # Fetch a batch and then add `variable` to the next batch.
                        dataframes.append(self._fetch_variables(variables))
                        variables.clear()
                        budget = URL_CHARACTER_LIMIT - (
                            _EXTRA_CHARS_PER_ITEM + len(variable)
                        )
                    variables.append(variable)
                if variables:
                    # Fetch the final batch.
                    dataframes.append(self._fetch_variables(variables))
                self._dataframe = pandas.concat(dataframes, axis=1).reset_index()
        return self._dataframe

    def _fetch_variables(self, variables):
        content = handle_error(
            self.get(
                self.link,
                params={"format": APACHE_ARROW_FILE_MIME_TYPE, "field": variables},
            )
        ).read()
        return deserialize_arrow(content)


def write_xarray_dataset(client_node, dataset, key=None):
    dataset_client = client_node.create_container(
        key=key, specs=[Spec("xarray_dataset")], metadata={"attrs": dataset.attrs}
    )
    for name in dataset.data_vars:
        data_array = dataset[name]
        dataset_client.write_array(
            data_array.data,
            key=name,
            metadata={"attrs": data_array.attrs},
            dims=data_array.dims,
            specs=[Spec("xarray_data_var")],
        )
    for name in dataset.coords:
        data_array = dataset[name]
        dataset_client.write_array(
            data_array.data,
            key=name,
            metadata={"attrs": data_array.attrs},
            dims=data_array.dims,
            specs=[Spec("xarray_coord")],
        )
    return dataset_client



import collections.abc
import itertools

import xarray

from ..structures.core import Spec
from .array import ArrayAdapter
from .mapping import MapAdapter


class DatasetAdapter(MapAdapter):
    """
    Wrap an xarray.Dataset
    """

    @classmethod
    def from_dataset(cls, dataset, *, specs=None, access_policy=None):
        mapping = _DatasetMap(dataset)
        return cls(
            mapping,
            metadata={"attrs": dataset.attrs},
            specs=specs,
            access_policy=access_policy,
        )

    def __init__(self, mapping, *args, specs=None, access_policy=None, **kwargs):
        if isinstance(mapping, xarray.Dataset):
            raise TypeError(
                "Use DatasetAdapter.from_dataset(...), not DatasetAdapter(...)."
            )
        specs = specs or []
        specs.append(Spec("xarray_dataset"))
        super().__init__(
            mapping, *args, specs=specs, access_policy=access_policy, **kwargs
        )

    def inlined_contents_enabled(self, depth):
        # Tell the server to in-line the description of each array
        # (i.e. data_vars and coords) to avoid latency of a second
        # request.
        return True


class _DatasetMap(collections.abc.Mapping):
    def __init__(self, dataset):
        self._dataset = dataset

    def __len__(self):
        return len(self._dataset.data_vars) + len(self._dataset.coords)

    def __iter__(self):
        yield from itertools.chain(self._dataset.data_vars, self._dataset.coords)

    def __getitem__(self, key):
        data_array = self._dataset[key]
        if key in self._dataset.coords:
            spec = Spec("xarray_coord")
        else:
            spec = Spec("xarray_data_var")
        return ArrayAdapter.from_array(
            data_array.data,
            metadata={"attrs": data_array.attrs},
            dims=data_array.dims,
            specs=[spec],
        )



import io
import json
from pathlib import Path

import numpy
import pandas
import pytest
import xarray

from ..adapters.array import ArrayAdapter
from ..adapters.dataframe import DataFrameAdapter
from ..adapters.mapping import MapAdapter
from ..adapters.xarray import DatasetAdapter
from ..client import Context, from_context
from ..client.utils import ClientError
from ..server.app import build_app

data = numpy.random.random((10, 10))
temp = 15 + 8 * numpy.random.randn(2, 2, 3)
precip = 10 * numpy.random.rand(2, 2, 3)
lon = [[-99.83, -99.32], [-99.79, -99.23]]
lat = [[42.25, 42.21], [42.63, 42.59]]
tree = MapAdapter(
    {
        "A": ArrayAdapter.from_array(numpy.random.random((100, 100))),
        "B": ArrayAdapter.from_array(numpy.random.random((100, 100, 100))),
        "C": DataFrameAdapter.from_pandas(
            pandas.DataFrame(
                {
                    "x": 1 * numpy.random.random(100),
                    "y": 2 * numpy.random.random(100),
                    "z": 3 * numpy.random.random(100),
                }
            ),
            npartitions=3,
        ),
        "empty_table": DataFrameAdapter.from_pandas(
            pandas.DataFrame({"A": []}), npartitions=1
        ),
        "structured_data": MapAdapter(
            {
                "pets": ArrayAdapter.from_array(
                    numpy.array(
                        [("Rex", 9, 81.0), ("Fido", 3, 27.0)],
                        dtype=[("name", "U10"), ("age", "i4"), ("weight", "f4")],
                    )
                ),
                "weather": DatasetAdapter.from_dataset(
                    xarray.Dataset(
                        {
                            "temperature": (["x", "y", "time"], temp),
                            "precipitation": (["x", "y", "time"], precip),
                        },
                        coords={
                            "lon": (["x", "y"], lon),
                            "lat": (["x", "y"], lat),
                            "time": [1, 2, 3],  # using ints here so HDF5 can export
                        },
                    )
                ),
            }
        ),
    }
)


@pytest.fixture(scope="module")
def client():
    app = build_app(tree)
    with Context.from_app(app) as context:
        client = from_context(context)
        yield client


# We test a little bit of actual file export, using the tmpdir fixture,
# but we mostly export for a buffer in memory because disk access
# can be very cloud on cloud CI VMs.


@pytest.mark.parametrize("filename", ["numbers.csv", "image.png", "image.tiff"])
def test_export_2d_array(client, filename, tmpdir):
    client["A"].export(Path(tmpdir, filename))


@pytest.mark.parametrize("filename", ["numbers.csv", "spreadsheet.xlsx"])
def test_export_table(client, filename, tmpdir):
    client["C"].export(Path(tmpdir, filename))


def test_streaming_export(client):
    "The application/json-seq format is streamed via a generator."
    buffer = io.BytesIO()
    client["C"].export(buffer, format="application/json-seq")
    # Verify that output is valid newline-delimited JSON.
    buffer.seek(0)
    lines = buffer.read().decode().splitlines()
    assert len(lines) == 100
    for line in lines:
        json.loads(line)


def test_streaming_export_empty(client):
    "The application/json-seq format is streamed via a generator."
    buffer = io.BytesIO()
    client["empty_table"].export(buffer, format="application/json-seq")
    buffer.seek(0)
    assert buffer.read() == b""


def test_export_weather_data_var(client, tmpdir):
    buffer = io.BytesIO()
    client["structured_data"]["weather"]["temperature"].export(
        buffer, slice=(0,), format="text/csv"
    )


def test_export_weather_all(client):
    buffer = io.BytesIO()
    client["structured_data"]["weather"].export(buffer, format="application/x-hdf5")


def test_serialization_error_hdf5_metadata(client):
    tree = MapAdapter(
        {
            "good": MapAdapter({}, metadata={"a": 1}),
            "bad": MapAdapter({}, metadata={"a": {"b": 1}}),
        }
    )
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        buffer = io.BytesIO()
        client["good"].export(buffer, format="application/x-hdf5")
        with pytest.raises(ClientError, match="contains types or structure"):
            client["bad"].export(buffer, format="application/x-hdf5")


def test_path_as_Path_or_string(client, tmpdir):
    client["A"].export(Path(tmpdir, "test_path_as_path.txt"))
    client["A"].export(str(Path(tmpdir, "test_path_as_str.txt")))


def test_formats(client):
    client.formats
    client["A"].formats



import random
import string
import uuid

import numpy
import pytest

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..client import Context, from_context
from ..server.app import build_app

sorted_letters = list(string.ascii_lowercase)[:10]
repeated_letters = ["a", "b"] * 5
sorted_numbers = list(range(10))
letters = sorted_letters.copy()
numbers = sorted_numbers.copy()
seed = random.Random(0)  # deterministic
seed.shuffle(letters)
seed.shuffle(numbers)

tree = MapAdapter(
    {
        str(uuid.uuid4()): ArrayAdapter.from_array(
            numpy.arange(10),
            metadata={
                "letter": letter,
                "number": number,
                "repeated_letter": repeated_letter,
            },
        )
        for letter, number, repeated_letter in zip(letters, numbers, repeated_letters)
    }
)


@pytest.fixture(scope="module")
def client():
    app = build_app(tree)
    with Context.from_app(app) as context:
        client = from_context(context)
        yield client


@pytest.mark.parametrize(
    "key, sorted_list",
    [
        ("letter", sorted_letters),
        ("number", sorted_numbers),
    ],
)
def test_sort(client, key, sorted_list):
    unsorted = [node.metadata[key] for node in client.values()]
    assert unsorted != sorted_list
    sorted_ascending = [node.metadata[key] for node in client.sort((key, 1)).values()]
    assert sorted_ascending == sorted_list
    sorted_descending = [node.metadata[key] for node in client.sort((key, -1)).values()]
    assert sorted_descending == list(reversed(sorted_list))


def test_sort_two_columns(client):
    # Sort by (repeated) letter, then by number.
    client_sorted = client.sort(("repeated_letter", 1), ("number", 1))
    letters_ = [node.metadata["repeated_letter"] for node in client_sorted.values()]
    numbers_ = [node.metadata["number"] for node in client_sorted.values()]
    # Letters are sorted.
    assert letters_ == ["a"] * 5 + ["b"] * 5
    # Numbers *within* each block of letters are sorted
    assert sorted(numbers_[:5]) == numbers_[:5]
    assert sorted(numbers_[5:]) == numbers_[5:]
    # but not sorted overall.
    assert not sorted(numbers_) == numbers_


def test_sort_sparse():
    """
    Sort where the key only present on some nodes.
    """
    tree = MapAdapter(
        {
            "yes1": ArrayAdapter.from_array(numpy.arange(10), metadata={"stuff": "a"}),
            "no1": ArrayAdapter.from_array(numpy.arange(10), metadata={}),
            "yes2": ArrayAdapter.from_array(numpy.arange(10), metadata={"stuff": "b"}),
            "no2": ArrayAdapter.from_array(numpy.arange(10), metadata={}),
        }
    )
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        client_sorted = client.sort(("stuff", 1))
        assert list(client_sorted)[:2] == ["yes1", "yes2"]


def test_sort_missing():
    """
    Sort where the key not present on any node.
    """
    tree = MapAdapter(
        {
            "no1": ArrayAdapter.from_array(numpy.arange(10), metadata={}),
            "no2": ArrayAdapter.from_array(numpy.arange(10), metadata={}),
        }
    )
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        client_sorted = client.sort(("stuff", 1))
        list(client_sorted)  # Just check for no errors.



import asyncio

from tiled._tests.test_catalog import temp_postgres
from tiled.catalog.adapter import Adapter
from tiled.catalog.explain import record_explanations
from tiled.queries import Key


async def test(a):
    for i in range(100):
        await a.create_container(
            metadata={
                "number": i,
                "number_as_string": str(i),
                "nested": {"number": i, "number_as_string": str(i)},
            },
            specs=[],
            structure_family="array",
        )
    # await a.create_metadata_index("nested_number_as_string", "nested.number_as_string")
    # await a.create_metadata_index("nested_number", "nested.number")
    with record_explanations() as e:
        results = await a.search(Key("number_as_string") == "3").keys_slice(0, 5, 1)
        assert len(results) == 1
        results = await a.search(Key("number") == 3).keys_slice(0, 5, 1)
        assert len(results) == 1
        results = await a.search(Key("nested.number_as_string") == "3").keys_slice(
            0, 5, 1
        )
        assert len(results) == 1
        results = await a.search(Key("nested.number") == 3).keys_slice(0, 5, 1)
        assert len(results) == 1
    print(e)


async def test_sqlite():
    async with Adapter.async_in_memory(echo=True) as a:
        await test(a)


async def test_postgres():
    async with temp_postgres(
        "postgresql+asyncpg://postgres:secret@localhost:5432", echo=True
    ) as a:
        await test(a)


# asyncio.run(test_postgres())
asyncio.run(test_sqlite())



import collections
import collections.abc

import httpx

from ..utils import import_object, prepend_to_sys_path
from .container import DEFAULT_STRUCTURE_CLIENT_DISPATCH, Container
from .context import DEFAULT_TIMEOUT_PARAMS, UNSET, Context
from .utils import MSGPACK_MIME_TYPE, ClientError, client_for_item, handle_error


def from_uri(
    uri,
    structure_clients="numpy",
    *,
    cache=UNSET,
    username=UNSET,
    auth_provider=UNSET,
    api_key=None,
    verify=True,
    prompt_for_reauthentication=UNSET,
    headers=None,
    timeout=None,
):
    """
    Connect to a Node on a local or remote server.

    Parameters
    ----------
    uri : str
        e.g. "http://localhost:8000"
    structure_clients : str or dict, optional
        Use "dask" for delayed data loading and "numpy" for immediate
        in-memory structures (e.g. normal numpy arrays, pandas
        DataFrames). For advanced use, provide dict mapping a
        structure_family or a spec to a client object.
    cache : Cache, optional
    username : str, optional
        Username for authenticated access. If UNSET, use default if available
        (typically, the most recently used).
    auth_provider : str, optional
        Name of an authentication provider. IF UNSET, use default if available
        (typically, the most recently used). If None and the server supports
        multiple provides, the user will be interactively prompted to
        choose from a list.
    api_key : str, optional
        API key based authentication. Cannot mix with username/auth_provider.
    verify : bool, optional
        Verify SSL certifications. True by default. False is insecure,
        intended for development and testing only.
    prompt_for_reauthentication : bool, optional
        If True, prompt interactively for credentials if needed. If False,
        raise an error. By default, attempt to detect whether terminal is
        interactive (is a TTY).
    headers : dict, optional
        Extra HTTP headers.
    timeout : httpx.Timeout, optional
        If None, use Tiled default settings.
        (To disable timeouts, use httpx.Timeout(None)).
    """
    context, node_path_parts = Context.from_any_uri(
        uri,
        api_key=api_key,
        cache=cache,
        headers=headers,
        timeout=timeout,
        verify=verify,
    )
    return from_context(
        context,
        structure_clients=structure_clients,
        prompt_for_reauthentication=prompt_for_reauthentication,
        username=username,
        auth_provider=auth_provider,
        node_path_parts=node_path_parts,
    )


def from_context(
    context,
    structure_clients="numpy",
    prompt_for_reauthentication=UNSET,
    username=UNSET,
    auth_provider=UNSET,
    node_path_parts=None,
):
    """
    Advanced: Connect to a Node using a custom instance of httpx.Client or httpx.AsyncClient.

    Parameters
    ----------
    context : tiled.client.context.Context
    structure_clients : str or dict, optional
        Use "dask" for delayed data loading and "numpy" for immediate
        in-memory structures (e.g. normal numpy arrays, pandas
        DataFrames). For advanced use, provide dict mapping a
        structure_family or a spec to a client object.
    prompt_for_reauthentication : bool, optional
        If True, prompt interactively for credentials if needed. If False,
        raise an error. By default, attempt to detect whether terminal is
        interactive (is a TTY).
    """
    if (username is not UNSET) or (auth_provider is not UNSET):
        if context.api_key is not None:
            raise ValueError("Use api_key or username/auth_provider, not both.")
    node_path_parts = node_path_parts or []
    # Do entrypoint discovery if it hasn't yet been done.
    if Container.STRUCTURE_CLIENTS_FROM_ENTRYPOINTS is None:
        Container.discover_clients_from_entrypoints()
    # Interpret structure_clients="numpy" and structure_clients="dask" shortcuts.
    if isinstance(structure_clients, str):
        structure_clients = DEFAULT_STRUCTURE_CLIENT_DISPATCH[structure_clients]
    if (
        (context.api_key is None)
        and context.server_info["authentication"]["required"]
        and (not context.server_info["authentication"]["providers"])
    ):
        raise RuntimeError(
            """This server requires API key authentication.
Set an api_key as in:

>>> c = from_uri("...", api_key="...")
"""
        )
    if username is not UNSET:
        context.authenticate(
            username=username,
            provider=auth_provider,
            prompt_for_reauthentication=prompt_for_reauthentication,
        )
    # Context ensures that context.api_uri has a trailing slash.
    item_uri = f"{context.api_uri}metadata/{'/'.join(node_path_parts)}"
    try:
        content = handle_error(
            context.http_client.get(item_uri, headers={"Accept": MSGPACK_MIME_TYPE})
        ).json()
    except ClientError as err:
        if (
            (err.response.status_code == 401)
            and (context.api_key is None)
            and (context.http_client.auth is None)
        ):
            context.authenticate()
            content = handle_error(
                context.http_client.get(item_uri, headers={"Accept": MSGPACK_MIME_TYPE})
            ).json()
        else:
            raise
    item = content["data"]
    return client_for_item(context, structure_clients, item)


def from_profile(name, structure_clients=None, **kwargs):
    """
    Build a Node based a 'profile' (a named configuration).

    List available profiles and the source filepaths from Python like:

    >>> from tiled.profiles import list_profiles
    >>> list_profiles()

    or from a CLI like:

    $ tiled profile list

    Or show the file contents like:

    >>> from tiled.profiles import load_profiles
    >>> load_profiles()

    or from a CLI like:

    $ tiled profile show PROFILE_NAME

    Any additional parameters override profile content. See from_uri for details.
    """
    # We accept structure_clients as a separate parameter so that it
    # may be invoked positionally, as in from_profile("...", "dask").
    from ..profiles import ProfileNotFound, load_profiles, paths

    profiles = load_profiles()
    try:
        filepath, profile_content = profiles[name]
    except KeyError as err:
        raise ProfileNotFound(
            f"Profile {name!r} not found. Found profiles {list(profiles)} "
            f"from directories {paths}."
        ) from err
    merged = {**profile_content, **kwargs}
    if structure_clients is not None:
        merged["structure_clients"] = structure_clients
    cache_config = merged.pop("cache", None)
    if cache_config is not None:
        from tiled.client.cache import Cache

        if isinstance(cache_config, collections.abc.Mapping):
            # All necessary validation has already been performed
            # in load_profiles().
            cache = Cache(**cache_config)
        else:
            # Interpret this as a Cache object passed in directly.
            cache = cache_config
        merged["cache"] = cache
    timeout_config = merged.pop("timeout", None)
    if timeout_config is not None:
        if isinstance(timeout_config, httpx.Timeout):
            timeout = timeout_config
        else:
            timeout_params = DEFAULT_TIMEOUT_PARAMS.copy()
            timeout_params.update(timeout_config)
            timeout = httpx.Timeout(**timeout_params)
        merged["timeout"] = timeout
    # Below, we may convert importable strings like
    # "package.module:obj" to live objects. Include the profile's
    # source directory in the import path, temporarily.
    with prepend_to_sys_path(filepath.parent):
        structure_clients_ = merged.pop("structure_clients", None)
        if structure_clients_ is not None:
            if isinstance(structure_clients_, str):
                # Nothing to do.
                merged["structure_clients"] = structure_clients_
            else:
                # This is a dict mapping structure families like "array" and "dataframe"
                # to values. The values may be client objects or importable strings.
                result = {}
                for key, value in structure_clients_.items():
                    if isinstance(value, str):
                        class_ = import_object(value, accept_live_object=True)
                    else:
                        class_ = value
                    result[key] = class_
                merged["structure_clients"] = result
    if "direct" in merged:
        # The profile specifies the server in-line.
        # Create an app and use it directly via ASGI.
        import jsonschema

        from ..config import ConfigError, schema
        from ..server.app import build_app_from_config

        config = merged.pop("direct", None)
        try:
            jsonschema.validate(instance=config, schema=schema())
        except jsonschema.ValidationError as err:
            msg = err.args[0]
            raise ConfigError(
                f"ValidationError while parsing configuration file {filepath}: {msg}"
            ) from err
        context = Context.from_app(
            build_app_from_config(config, source_filepath=filepath),
        )
        return from_context(context, **merged)
    else:
        return from_uri(**merged)



import warnings

warnings.warn(
    """The module 'tiled.client.node' has been moved to 'tiled.client.container' and
the object 'Node' has been renamed 'Container'.""",
    DeprecationWarning,
)
from .container import *  # noqa

Node = Container  # noqa



import string

import numpy
import pytest

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..client import Context, from_context
from ..server.app import build_app

keys = list(string.ascii_lowercase)
tree = MapAdapter(
    {
        letter: ArrayAdapter.from_array(number * numpy.ones(10))
        for letter, number in zip(keys, range(26))
    }
)
empty_tree = MapAdapter({})


@pytest.fixture(scope="module")
def client():
    app = build_app(tree)
    with Context.from_app(app) as context:
        client = from_context(context)
        yield client


@pytest.fixture(scope="module")
def empty_client():
    app = build_app(empty_tree)
    with Context.from_app(app) as context:
        client = from_context(context)
        yield client


def test_indexers(client, empty_client):
    assert client.keys()[0] == client.keys().first() == keys[0] == "a"
    assert client.keys()[-1] == client.keys().last() == keys[-1] == "z"
    assert client.keys()[:3] == client.keys().head(3) == keys[:3] == list("abc")
    assert client.keys()[:1] == client.keys().head(1) == keys[:1] == list("a")
    assert client.keys()[1:3] == keys[1:3] == list("bc")
    assert client.keys()[-3::-1] == keys[-3::-1]  # ["x", "w", "v", ...]
    assert client.keys()[-1:-5:-1] == keys[-1:-5:-1] == list("zyxw")
    assert client.keys().tail(4) == list(reversed(keys[-1:-5:-1])) == list("wxyz")
    assert client.keys()[:] == list(client) == keys
    # Slice beyond length.
    assert 100 > len(client.keys())
    assert client.keys()[:100] == client.keys().head(100) == keys[:100] == keys[:]
    assert empty_client.keys().head() == []
    assert empty_client.keys().tail() == []
    # Test out of bounds
    with pytest.raises(IndexError):
        client.keys()[len(keys)]
    with pytest.raises(IndexError):
        client.keys()[-len(keys) - 1]
    with pytest.raises(IndexError):
        empty_client.keys().first()
    with pytest.raises(IndexError):
        empty_client.keys().last()
    # These should be in bounds and should not raise.
    client.keys()[len(keys) - 1]
    client.keys()[-len(keys)]

    # smoke test values()
    client.values().first()
    client.values().last()
    client.values().head()
    client.values().tail()
    client.values()[1:3]

    # smoke test items()
    client.items().first()
    client.items().last()
    client.items().head()
    client.items().tail()
    client.items()[1:3]


def test_deprecated_indexer_accessors(client):
    with pytest.warns(DeprecationWarning):
        assert client.keys_indexer[:3] == keys[:3] == list("abc")
    # smoke test the others
    with pytest.warns(DeprecationWarning):
        client.values_indexer[:3]
    with pytest.warns(DeprecationWarning):
        client.items_indexer[:3]



"""
A directory containing awkward buffers, one file per form key.
"""
import collections.abc

import awkward.forms

from ..structures.core import StructureFamily
from ..utils import path_from_uri
from .awkward import AwkwardAdapter


class DirectoryContainer(collections.abc.MutableMapping):
    def __init__(self, directory, form):
        self.directory = directory
        self.form = form

    def __getitem__(self, form_key):
        with open(self.directory / form_key, "rb") as file:
            return file.read()

    def __setitem__(self, form_key, value):
        with open(self.directory / form_key, "wb") as file:
            file.write(value)

    def __delitem__(self, form_key):
        (self.directory / form_key).unlink(missing_ok=True)

    def __iter__(self):
        yield from self.form.expected_from_buffers()

    def __len__(self):
        return len(self.form.expected_from_buffers())


class AwkwardBuffersAdapter(AwkwardAdapter):
    structure_family = StructureFamily.awkward

    @classmethod
    def init_storage(cls, data_uri, structure):
        from ..server.schemas import Asset

        directory = path_from_uri(data_uri)
        directory.mkdir(parents=True, exist_ok=True)
        return [Asset(data_uri=data_uri, is_directory=True, parameter="data_uri")]

    @classmethod
    def from_directory(
        cls,
        data_uri,
        structure,
        metadata=None,
        specs=None,
        access_policy=None,
    ):
        form = awkward.forms.from_dict(structure.form)
        directory = path_from_uri(data_uri)
        if not directory.is_dir():
            raise ValueError(f"Not a directory: {directory}")
        container = DirectoryContainer(directory, form)
        return cls(
            container,
            structure=structure,
            metadata=metadata,
            specs=specs,
            access_policy=access_policy,
        )



import numpy
import pytest

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..catalog import in_memory
from ..client import Context, from_context
from ..queries import FullText, Key
from ..server.app import build_app

tree = MapAdapter(
    {
        "a": ArrayAdapter.from_array(
            numpy.arange(10), metadata={"apple": "red", "animal": "dog"}
        ),
        "b": ArrayAdapter.from_array(
            numpy.arange(10), metadata={"banana": "yellow", "animal": "dog"}
        ),
        "c": ArrayAdapter.from_array(
            numpy.arange(10), metadata={"cantalope": "orange", "animal": "cat"}
        ),
    }
)


@pytest.fixture(scope="module")
def map_client():
    app = build_app(tree)
    with Context.from_app(app) as context:
        client = from_context(context)
        yield client


@pytest.fixture
def catalog_client(tmpdir):
    catalog = in_memory(writable_storage=str(tmpdir))
    app = build_app(catalog)
    with Context.from_app(app) as context:
        client = from_context(context)
        # Upload arrays and metadata from tree into client so that it has matching content.
        for key, value in tree.items():
            client.write_array(value.read(), metadata=value.metadata(), key=key)
        yield client


@pytest.fixture(
    params=("map_client", "catalog_client"),
)
def client(request: pytest.FixtureRequest):
    yield request.getfixturevalue(request.param)


@pytest.mark.parametrize(
    "key, value, expected_keys",
    [
        ("apple", "red", ["a"]),
        ("banana", "yellow", ["b"]),
        ("cantalope", "orange", ["c"]),
        ("animal", "dog", ["a", "b"]),
        ("animal", "cat", ["c"]),
    ],
)
def test_search(client, key, value, expected_keys):
    query = Key(key)
    results = client.search(query == value)
    assert list(results) == expected_keys


def test_compound_search(client):
    results = client.search(Key("animal") == "dog").search(Key("banana") == "yellow")
    assert list(results) == ["b"]


def test_indexing_over_search(client):
    results = client.search(Key("animal") == "dog")
    assert dict(results["a"].metadata) == tree["a"].metadata()


def test_key_into_results(client):
    results = client.search(Key("animal") == "dog")
    assert "apple" in results["a"].metadata
    assert "banana" in results["b"].metadata
    assert "c" not in results  # This *is* in the tree but not among the results.


def test_compound_key_into_results():
    nested_tree = MapAdapter(
        {
            "i": MapAdapter({"X": tree}, metadata={"temp": "hot"}),
            "j": MapAdapter({}, metadata={"temp": "cold"}),
        }
    )
    app = build_app(nested_tree)
    with Context.from_app(app) as context:
        client = from_context(context)
        result = client.search(FullText("hot"))["i", "X", "a"]
        assert "apple" in result.metadata



import asyncio
import string
import subprocess
import sys
from contextlib import nullcontext

import numpy
import pytest
import pytest_asyncio

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..catalog import from_uri, in_memory
from ..client import Context, from_context
from ..queries import (
    Comparison,
    Contains,
    Eq,
    FullText,
    In,
    Key,
    KeysFilter,
    NotEq,
    NotIn,
    Regex,
    SpecsQuery,
    StructureFamilyQuery,
)
from ..server.app import build_app
from .conftest import TILED_TEST_POSTGRESQL_URI
from .utils import fail_with_status_code, temp_postgres

keys = list(string.ascii_lowercase)
mapping = {
    letter: ArrayAdapter.from_array(
        number * numpy.ones(10),
        metadata={"letter": letter, "number": number},
        specs=[letter],
    )
    for letter, number in zip(keys, range(26))
}
mapping["does_contain_z"] = ArrayAdapter.from_array(
    numpy.ones(10), metadata={"letters": list(string.ascii_lowercase)}
)
mapping["does_not_contain_z"] = ArrayAdapter.from_array(
    numpy.ones(10), metadata={"letters": list(string.ascii_lowercase[:-1])}
)
mapping["full_text_test_case"] = ArrayAdapter.from_array(
    numpy.ones(10), metadata={"color": "purple"}
)

mapping["specs_foo_bar"] = ArrayAdapter.from_array(numpy.ones(10), specs=["foo", "bar"])
mapping["specs_foo_bar_baz"] = ArrayAdapter.from_array(
    numpy.ones(10), specs=["foo", "bar", "baz"]
)


@pytest.fixture(scope="module")
def event_loop():
    # https://stackoverflow.com/a/56238383/1221924
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest_asyncio.fixture(scope="module", params=["map", "sqlite", "postgresql"])
async def client(request, tmpdir_module):
    if request.param == "map":
        tree = MapAdapter(mapping, metadata={"backend": request.param})
        app = build_app(tree)
        with Context.from_app(app) as context:
            client = from_context(context)
            yield client
    elif request.param == "sqlite":
        tree = in_memory(
            writable_storage=tmpdir_module / "sqlite",
            metadata={"backend": request.param},
        )
        app = build_app(tree)
        with Context.from_app(app) as context:
            client = from_context(context)
            for k, v in mapping.items():
                client.write_array(
                    v.read(), key=k, metadata=dict(v.metadata()), specs=v.specs
                )
            yield client
    elif request.param == "postgresql":
        if not TILED_TEST_POSTGRESQL_URI:
            raise pytest.skip("No TILED_TEST_POSTGRESQL_URI configured")
        # Create temporary database.
        async with temp_postgres(TILED_TEST_POSTGRESQL_URI) as uri_with_database_name:
            subprocess.run(
                [
                    sys.executable,
                    "-m",
                    "tiled",
                    "catalog",
                    "init",
                    uri_with_database_name,
                ],
                check=True,
                capture_output=True,
            )
            tree = from_uri(
                uri_with_database_name,
                writable_storage=str(tmpdir_module / "postgresql"),
                metadata={"backend": request.param},
            )
            app = build_app(tree)
            with Context.from_app(app) as context:
                client = from_context(context)
                # Write data into catalog.
                for k, v in mapping.items():
                    client.write_array(
                        v.read(),
                        key=k,
                        metadata=dict(v.metadata()),
                        specs=v.specs,
                    )
                yield client
    else:
        assert False


def test_key(client):
    "Binary operators with Key create query objects."
    assert (Key("a") == 1) == Eq("a", 1)
    assert (Key("a") != 2) == NotEq("a", 2)
    assert (Key("a") > 1) == Comparison("gt", "a", 1)
    assert (Key("a") < 1) == Comparison("lt", "a", 1)
    assert (Key("a") >= 1) == Comparison("ge", "a", 1)
    assert (Key("a") <= 1) == Comparison("le", "a", 1)


def test_eq(client):
    # Test encoding letters and ints.
    assert list(client.search(Eq("letter", "a"))) == ["a"]
    assert list(client.search(Eq("letter", "b"))) == ["b"]
    assert list(client.search(Eq("number", 0))) == ["a"]
    assert list(client.search(Eq("number", 1))) == ["b"]
    # Number is an int, not a string, so this should not match anything.
    assert list(client.search(Eq("number", "0"))) == []


def test_noteq(client):
    # Test encoding letters and ints.
    assert list(client.search(NotEq("letter", "a"))) != ["a"]
    assert list(client.search(NotEq("letter", "b"))) != ["b"]
    assert list(client.search(NotEq("number", 0))) != ["a"]
    assert list(client.search(NotEq("number", 1))) != ["b"]


def test_comparison(client):
    assert list(client.search(Comparison("gt", "number", 24))) == ["z"]
    assert list(client.search(Comparison("ge", "number", 24))) == ["y", "z"]
    assert list(client.search(Comparison("lt", "number", 1))) == ["a"]
    assert list(client.search(Comparison("le", "number", 1))) == ["a", "b"]


def test_contains(client):
    assert list(client.search(Contains("letters", "z"))) == ["does_contain_z"]


def test_full_text(client):
    if client.metadata["backend"] in {"sqlite"}:

        def cm():
            return fail_with_status_code(400)

    else:
        cm = nullcontext
    with cm():
        assert list(client.search(FullText("z"))) == ["z", "does_contain_z"]
        # plainto_tsquery fails to find certain words, weirdly, so it is a useful
        # test that we are using tsquery
        assert list(client.search(FullText("purple"))) == ["full_text_test_case"]


def test_regex(client):
    if client.metadata["backend"] in {"postgresql", "sqlite"}:

        def cm():
            return fail_with_status_code(400)

    else:
        cm = nullcontext
    with cm():
        assert list(client.search(Regex("letter", "^z$"))) == ["z"]
    with cm():
        assert (
            list(client.search(Regex("letter", "^Z$"))) == []
        )  # default case_sensitive=True
    with cm():
        assert list(client.search(Regex("letter", "^Z$", case_sensitive=False))) == [
            "z"
        ]
    with cm():
        assert list(client.search(Regex("letter", "^Z$", case_sensitive=True))) == []
    with cm():
        assert list(client.search(Regex("letter", "[a-c]"))) == ["a", "b", "c"]
    with cm():
        # Check that if the key is not a string it is ignored.
        assert list(client.search(Regex("letters", "anything"))) == []


def test_not_and_and_or(client):
    with pytest.raises(TypeError):
        not (Key("color") == "red")
    with pytest.raises(TypeError):
        (Key("color") == "red") and (Key("sample") == "Ni")
    with pytest.raises(TypeError):
        (Key("color") == "red") or (Key("sample") == "Ni")


@pytest.mark.parametrize(
    "query_values",
    [
        ["a", "k", "z"],
        ("a", "k", "z"),
        {"a", "k", "z"},
        {"a", "k", "z", "a", "z", "z"},
    ],
)
def test_in(client, query_values):
    assert sorted(list(client.search(In("letter", query_values)))) == [
        "a",
        "k",
        "z",
    ]


@pytest.mark.parametrize(
    "query_values",
    [
        ["a", "k", "z"],
        ("a", "k", "z"),
        {"a", "k", "z"},
        {"a", "k", "z", "a", "z", "z"},
    ],
)
def test_notin(client, query_values):
    # TODO: Postgres and SQlite ACTUALLY treat this query differently in external testing.
    # SQLite WILL NOT include fields that do not have the key, which is correct.
    # Postgres WILL include fields that do not have the key,
    # because by extension they do not have the value. Also correct. Why?
    assert sorted(list(client.search(NotIn("letter", query_values)))) == sorted(
        list(
            set(
                list(mapping.keys())
                if client.metadata["backend"] == "postgresql"
                else keys
            )
            - set(["a", "k", "z"])
        )
    )


@pytest.mark.parametrize(
    "include_values,exclude_values",
    [
        (["foo", "bar"], ["baz"]),
        (("foo", "bar"), ("baz",)),
        ({"foo", "bar"}, {"baz"}),
        ({"foo", "bar", "foo", "bar", "bar"}, {"baz", "baz", "baz"}),
    ],
)
def test_specs(client, include_values, exclude_values):
    assert list(
        client.search(SpecsQuery(include=include_values, exclude=exclude_values))
    ) == ["specs_foo_bar"]


def test_structure_families(client):
    with pytest.raises(ValueError):
        StructureFamilyQuery("foo")

    assert set(client.search(StructureFamilyQuery("array"))) == set(mapping)


def test_keys_filter(client):
    expected = ["a", "b", "c"]
    results = client.search(KeysFilter(keys=expected))
    assert set(results) == set(expected)



"""
This module handles server configuration.

See profiles.py for client configuration.
"""
import copy
import os
from collections import defaultdict
from datetime import timedelta
from functools import lru_cache
from pathlib import Path

import jsonschema

from .adapters.mapping import MapAdapter
from .media_type_registration import (
    compression_registry as default_compression_registry,
)
from .media_type_registration import (
    serialization_registry as default_serialization_registry,
)
from .query_registration import query_registry as default_query_registry
from .utils import import_object, parse, prepend_to_sys_path
from .validation_registration import validation_registry as default_validation_registry


@lru_cache(maxsize=1)
def schema():
    "Load the schema for service-side configuration."
    import yaml

    here = Path(__file__).parent.absolute()
    schema_path = here / "config_schemas" / "service_configuration.yml"
    with open(schema_path, "r") as file:
        return yaml.safe_load(file)


def construct_build_app_kwargs(
    config,
    *,
    source_filepath=None,
    query_registry=None,
    compression_registry=None,
    serialization_registry=None,
    validation_registry=None,
):
    """
    Given parsed configuration, construct arguments for build_app(...).

    The parameters query_registry, compression_registry, and
    serialization_registry are used by the tests to inject separate registry
    instances. By default, the singleton global instances of these registries
    and used (and modified).
    """
    config = copy.deepcopy(config)  # Avoid mutating input.
    startup_tasks = []
    shutdown_tasks = []
    background_tasks = []
    if query_registry is None:
        query_registry = default_query_registry
    if serialization_registry is None:
        serialization_registry = default_serialization_registry
    if compression_registry is None:
        compression_registry = default_compression_registry
    if validation_registry is None:
        validation_registry = default_validation_registry
    sys_path_additions = []
    if source_filepath:
        if os.path.isdir(source_filepath):
            directory = source_filepath
        else:
            directory = os.path.dirname(source_filepath)
        sys_path_additions.append(directory)
    with prepend_to_sys_path(*sys_path_additions):
        auth_spec = config.get("authentication", {}) or {}
        for age in ["refresh_token_max_age", "session_max_age", "access_token_max_age"]:
            if age in auth_spec:
                auth_spec[age] = timedelta(seconds=auth_spec[age])
        root_access_control = config.get("access_control", {}) or {}
        auth_aliases = {}  # TODO Enable entrypoint as alias for authenticator_class?
        providers = list(auth_spec.get("providers", []))
        provider_names = [p["provider"] for p in providers]
        if len(set(provider_names)) != len(provider_names):
            raise ValueError(
                "The names given for 'provider' must be unique. "
                f"Found duplicates in {providers}"
            )
        for i, authenticator in enumerate(providers):
            import_path = auth_aliases.get(
                authenticator["authenticator"], authenticator["authenticator"]
            )
            authenticator_class = import_object(import_path, accept_live_object=True)
            authenticator = authenticator_class(**authenticator.get("args", {}))
            # Replace "package.module:Object" with live instance.
            auth_spec["providers"][i]["authenticator"] = authenticator
        if root_access_control.get("access_policy") is not None:
            root_policy_import_path = root_access_control["access_policy"]
            root_policy_class = import_object(
                root_policy_import_path, accept_live_object=True
            )
            root_access_policy = root_policy_class(
                **root_access_control.get("args", {})
            )
        else:
            root_access_policy = None
        # TODO Enable entrypoint to extend aliases?
        tree_aliases = {
            "catalog": "tiled.catalog:from_uri",
        }
        trees = {}
        for item in config.get("trees", []):
            access_control = item.get("access_control", {}) or {}
            if access_control.get("access_policy") is not None:
                policy_import_path = access_control["access_policy"]
                policy_class = import_object(
                    policy_import_path, accept_live_object=True
                )
                access_policy = policy_class(**access_control.get("args", {}))
            else:
                access_policy = None
            segments = tuple(segment for segment in item["path"].split("/") if segment)
            tree_spec = item["tree"]
            if isinstance(tree_spec, str) and tree_spec == "files":
                raise Exception(
                    """The way that tiled serves files has changed.

See documentation section "Serve a Directory of Files"."""
                )
            import_path = tree_aliases.get(tree_spec, tree_spec)
            obj = import_object(import_path, accept_live_object=True)
            if (("args" in item) or ("access_policy" in item)) and (not callable(obj)):
                raise ValueError(
                    f"Object imported from {import_path} cannot take args. "
                    "It is not callable."
                )
            if callable(obj):
                # Interpret obj as a tree *factory*.
                args = {}
                args.update(item.get("args", {}))
                if access_policy is not None:
                    args["access_policy"] = access_policy
                tree = obj(**args)
            else:
                # Interpret obj as a tree *instance*.
                if access_policy is not None:
                    raise ValueError(
                        f"Cannot apply access_policy to object {obj} which is already instantiated."
                    )
                tree = obj
            if segments in trees:
                raise ValueError(f"The path {'/'.join(segments)} was specified twice.")
            trees[segments] = tree
            startup_tasks.extend(getattr(tree, "startup_tasks", []))
            shutdown_tasks.extend(getattr(tree, "shutdown_tasks", []))
            background_tasks.extend(getattr(tree, "background_tasks", []))
        if not len(trees):
            raise ValueError("Configuration contains no trees")
        if list(trees) == [()]:
            # Simple case: there is one tree, served at the root path /.
            root_tree = tree
            if root_access_policy is not None:
                tree.access_policy = root_access_policy
        else:
            # There are one or more tree(s) to be served at
            # sub-paths. Merged them into one root MapAdapter.
            # Map path segments to dicts containing Adapters at that path.
            root_mapping = {}
            index = {(): root_mapping}
            include_routers = []
            for segments, tree in trees.items():
                for i in range(len(segments)):
                    if segments[:i] not in index:
                        mapping = {}
                        index[segments[:i]] = mapping
                        parent = index[segments[: i - 1]]
                        parent[segments[i - 1]] = MapAdapter(mapping)
                index[segments[:-1]][segments[-1]] = tree
                # Collect any custom routers.
                routers = getattr(tree, "include_routers", [])
                for router in routers:
                    if router not in include_routers:
                        include_routers.append(router)
            root_tree = MapAdapter(root_mapping, access_policy=root_access_policy)
            root_tree.include_routers.extend(include_routers)
        server_settings = {}
        if root_path := config.get("root_path", ""):
            server_settings["root_path"] = root_path
        server_settings["allow_origins"] = config.get("allow_origins")
        server_settings["object_cache"] = config.get("object_cache", {})
        server_settings["response_bytesize_limit"] = config.get(
            "response_bytesize_limit"
        )
        server_settings["database"] = config.get("database", {})
        server_settings["reject_undeclared_specs"] = config.get(
            "reject_undeclared_specs"
        )
        server_settings["metrics"] = config.get("metrics", {})
        for structure_family, values in config.get("media_types", {}).items():
            for media_type, import_path in values.items():
                serializer = import_object(import_path, accept_live_object=True)
                serialization_registry.register(
                    structure_family, media_type, serializer
                )
        for ext, media_type in config.get("file_extensions", {}).items():
            serialization_registry.register_alias(ext, media_type)

        for item in config.get("specs", []):
            if "validator" in item:
                validator = import_object(item["validator"])
            else:
                # no-op
                validator = _no_op_validator
            validation_registry.register(item["spec"], validator)

    # TODO Make compression_registry extensible via configuration.
    return {
        "tree": root_tree,
        "authentication": auth_spec,
        "server_settings": server_settings,
        "query_registry": query_registry,
        "serialization_registry": serialization_registry,
        "compression_registry": compression_registry,
        "validation_registry": validation_registry,
        "tasks": {
            "startup": startup_tasks,
            "shutdown": shutdown_tasks,
            "background": background_tasks,
        },
    }


def merge(configs):
    merged = {"trees": []}

    # These variables are used to produce error messages that point
    # to the relevant config file(s).
    authentication_config_source = None
    access_control_config_source = None
    uvicorn_config_source = None
    object_cache_config_source = None
    metrics_config_source = None
    database_config_source = None
    response_bytesize_limit_config_source = None
    allow_origins = []
    media_types = defaultdict(dict)
    specs = []
    reject_undeclared_specs_source = None
    file_extensions = {}
    paths = {}  # map each item's path to config file that specified it

    for filepath, config in configs.items():
        for structure_family, values in config.get("media_types", {}).items():
            media_types[structure_family].update(values)

        file_extensions.update(config.get("file_extensions", {}))
        allow_origins.extend(config.get("allow_origins", []))
        if "access_control" in config:
            if "access_control" in merged:
                raise ConfigError(
                    "access_control can only be specified in one file. "
                    f"It was found in both {access_control_config_source} and "
                    f"{filepath}"
                )
            access_control_config_source = filepath
            merged["access_control"] = config["access_control"]
        if "authentication" in config:
            if "authentication" in merged:
                raise ConfigError(
                    "authentication can only be specified in one file. "
                    f"It was found in both {authentication_config_source} and "
                    f"{filepath}"
                )
            authentication_config_source = filepath
            merged["authentication"] = config["authentication"]
        if "uvicorn" in config:
            if "uvicorn" in merged:
                raise ConfigError(
                    "uvicorn can only be specified in one file. "
                    f"It was found in both {uvicorn_config_source} and "
                    f"{filepath}"
                )
            uvicorn_config_source = filepath
            merged["uvicorn"] = config["uvicorn"]
        if "object_cache" in config:
            if "object_cache" in merged:
                raise ConfigError(
                    "object_cache can only be specified in one file. "
                    f"It was found in both {object_cache_config_source} and "
                    f"{filepath}"
                )
            object_cache_config_source = filepath
            merged["object_cache"] = config["object_cache"]
        if "response_bytesize_limit" in config:
            if "response_bytesize_limit" in merged:
                raise ConfigError(
                    "response_bytesize_limit can only be specified in one file. "
                    f"It was found in both {response_bytesize_limit_config_source} and "
                    f"{filepath}"
                )
            response_bytesize_limit_config_source = filepath
            merged["response_bytesize_limit"] = config["response_bytesize_limit"]
        if "metrics" in config:
            if "metrics" in merged:
                raise ConfigError(
                    "metrics can only be specified in one file. "
                    f"It was found in both {metrics_config_source} and "
                    f"{filepath}"
                )
            metrics_config_source = filepath
            merged["metrics"] = config["metrics"]
        if "database" in config:
            if "database" in merged:
                raise ConfigError(
                    "database configuration can only be specified in one file. "
                    f"It was found in both {database_config_source} and "
                    f"{filepath}"
                )
            database_config_source = filepath
            merged["database"] = config["database"]
        if "reject_undeclared_specs" in config:
            if "reject_undeclared_specs" in merged:
                raise ConfigError(
                    "'reject_undeclared_specs' can only be specified in one file. "
                    f"It was found in both {reject_undeclared_specs_source} and "
                    f"{filepath}"
                )
            reject_undeclared_specs_source = filepath
            merged["reject_undeclared_specs"] = config["reject_undeclared_specs"]
        for item in config.get("trees", []):
            if item["path"] in paths:
                msg = "A given path may be only be specified once."
                "The path {item['path']} was found twice in "
                if filepath == paths[item["path"]]:
                    msg += f"{filepath}."
                else:
                    msg += f"{filepath} and {paths[item['path']]}."
                raise ConfigError(msg)
            paths[item["path"]] = filepath
            merged["trees"].append(item)
        specs.extend(config.get("specs", []))
    merged["media_types"] = dict(media_types)  # convert from defaultdict
    merged["file_extensions"] = file_extensions
    merged["allow_origins"] = allow_origins
    merged["specs"] = specs
    return merged


def parse_configs(config_path):
    """
    Parse configuration file or directory of configuration files.

    If a directory is given, any files not ending in `.yml` or `.yaml` are
    ignored. Therefore, the directory may also contain a README file and
    supporting Python scripts with custom objects.
    """
    if isinstance(config_path, str):
        config_path = Path(config_path)
    if config_path.is_file():
        filepaths = [config_path]
    elif config_path.is_dir():
        filepaths = [
            fn for fn in config_path.iterdir() if fn.suffix in (".yml", ".yaml")
        ]
    elif not config_path.exists():
        raise ValueError(f"The config path {config_path!s} doesn't exist.")
    else:
        assert False, "It should be impossible to reach this line."

    parsed_configs = {}
    # The sorting here is just to make the order of the results deterministic.
    # There is *not* any sorting-based precedence applied.
    for filepath in sorted(filepaths):
        # Ignore hidden files and .py files.
        if (
            filepath.parts[-1].startswith(".")
            or filepath.suffix == ".py"
            or filepath.parts[-1] == "__pycache__"
        ):
            continue
        with open(filepath) as file:
            config = parse(file)
            try:
                jsonschema.validate(instance=config, schema=schema())
            except jsonschema.ValidationError as err:
                msg = err.args[0]
                raise ConfigError(
                    f"ValidationError while parsing configuration file {filepath}: {msg}"
                ) from err
            parsed_configs[filepath] = config

    merged_config = merge(parsed_configs)
    return merged_config


class ConfigError(ValueError):
    pass


def _no_op_validator(*args, **kwargs):
    return None



"""
Example using numpy structured dtypes

https://numpy.org/doc/stable/user/basics.rec.html

"""
import numpy

from tiled.adapters.array import ArrayAdapter
from tiled.adapters.mapping import MapAdapter

tree = MapAdapter(
    {
        "A": ArrayAdapter.from_array(
            numpy.array(
                [("Rex", 9, 81.0), ("Fido", 3, 27.0)],
                dtype=[("name", "U10"), ("age", "i4"), ("weight", "f4")],
            )
        )
    }
)



import numpy
import sparse
from ndindex import ndindex

from ..serialization.table import deserialize_arrow, serialize_arrow
from ..utils import APACHE_ARROW_FILE_MIME_TYPE
from .base import BaseClient
from .utils import export_util, handle_error, params_from_slice


class SparseClient(BaseClient):
    @property
    def dims(self):
        return self.structure().dims

    @property
    def shape(self):
        return self.structure().shape

    @property
    def chunks(self):
        return self.structure().chunks

    @property
    def ndim(self):
        return len(self.structure().shape)

    def __repr__(self):
        structure = self.structure()
        attrs = {"shape": structure.shape, "chunks": structure.chunks}
        if structure.dims:
            attrs["dims"] = structure.dims
        return (
            f"<{type(self).__name__}"
            + "".join(f" {k}={v}" for k, v in attrs.items())
            + ">"
        )

    def __array__(self, *args, **kwargs):
        return self.read().__array__(*args, **kwargs)

    def todense(self, *args, **kwargs):
        "Return a dense numpy array. May be large."
        return self.read().todense(*args, **kwargs)

    def read_block(self, block, slice=None):
        # Fetch the data as an Apache Arrow table
        # with columns named dim0, dim1, ..., dimN, data.
        structure = self.structure()
        params = params_from_slice(slice)
        params["block"] = ",".join(map(str, block))
        content = handle_error(
            self.context.http_client.get(
                self.item["links"]["block"],
                headers={"Accept": APACHE_ARROW_FILE_MIME_TYPE},
                params=params,
            )
        ).read()
        df = deserialize_arrow(content)
        original_shape = structure.shape
        if slice is not None:
            sliced_shape = ndindex(slice).newshape(original_shape)
        else:
            sliced_shape = original_shape
        ndim = len(sliced_shape)

        return sparse.COO(
            data=df["data"].values,
            coords=numpy.stack([df[f"dim{i}"].values for i in range(ndim)]),
            shape=sliced_shape,
        )

    def read(self, slice=None):
        # Fetch the data as an Apache Arrow table
        # with columns named dim0, dim1, ..., dimN, data.
        structure = self.structure()
        params = params_from_slice(slice)
        content = handle_error(
            self.context.http_client.get(
                self.item["links"]["full"],
                headers={"Accept": APACHE_ARROW_FILE_MIME_TYPE},
                params=params,
            )
        ).read()
        df = deserialize_arrow(content)
        original_shape = structure.shape
        if slice is not None:
            sliced_shape = ndindex(slice).newshape(original_shape)
        else:
            sliced_shape = original_shape
        ndim = len(sliced_shape)

        return sparse.COO(
            data=df["data"].values,
            coords=numpy.stack([df[f"dim{i}"].values for i in range(ndim)]),
            shape=sliced_shape,
        )

    def write(self, coords, data):
        import pandas

        d = {f"dim{i}": coords for i, coords in enumerate(coords)}
        d["data"] = data
        df = pandas.DataFrame(d)
        handle_error(
            self.context.http_client.put(
                self.item["links"]["full"],
                content=bytes(serialize_arrow(df, {})),
                headers={"Content-Type": APACHE_ARROW_FILE_MIME_TYPE},
            )
        )

    def write_block(self, coords, data, block):
        import pandas

        d = {f"dim{i}": coords for i, coords in enumerate(coords)}
        d["data"] = data
        df = pandas.DataFrame(d)
        handle_error(
            self.context.http_client.put(
                self.item["links"]["block"].format(*block),
                content=bytes(serialize_arrow(df, {})),
                headers={"Content-Type": APACHE_ARROW_FILE_MIME_TYPE},
            )
        )

    def __getitem__(self, slice):
        return self.read(slice)

    def export(self, filepath, *, format=None, slice=None):
        """
        Download data in some format and write to a file.

        Parameters
        ----------
        file: str or buffer
            Filepath or writeable buffer.
        format : str, optional
            If format is None and `file` is a filepath, the format is inferred
            from the name, like 'table.csv' implies format="text/csv". The format
            may be given as a file extension ("csv") or a media type ("text/csv").
        slice : List[slice], optional
            List of slice objects. A convenient way to generate these is shown
            in the examples.

        Examples
        --------

        Export all.

        >>> a.export("numbers.csv")

        Export an N-dimensional slice.

        >>> import numpy
        >>> a.export("numbers.csv", slice=numpy.s_[:10, 50:100])
        """
        params = params_from_slice(slice)
        return export_util(
            filepath,
            format,
            self.context.http_client.get,
            self.item["links"]["full"],
            params=params,
        )



"""
These objects express high-level queries and translate them (when possible)
into concrete queries for specific storage backends.

This intentionally only uses built-in dataclasses, not pydantic models.
"""
import inspect
from dataclasses import fields

from .utils import DictView, UnsupportedQueryType


class QueryRegistry:
    """
    Keep track of all known queries types, with names.

    When the server starts up, it uses this registry to populate the list of
    allowed URL query parameters to its /search route.

    There is a global instance of this, defined below. It is implemented as a
    class, rather than a module-scope singleton, for the sake of tests.
    """

    def __init__(self):
        self._name_to_query_type_type = {}
        self._query_type_to_name = {}

    @property
    def name_to_query_type(self):
        return DictView(self._name_to_query_type_type)

    @property
    def query_type_to_name(self):
        return DictView(self._query_type_to_name)

    def register(self, name=None, overwrite=False, must_revalidate=True):
        """
        Register a new type of query.
        """
        if "___" in name:
            raise Exception("Names must not contain triple underscores ('___').")
            # Why? This would create ambiguity in the server's handling of
            # search requests. Route signature parameters are named like
            # "filter___{name}___{field}".

        def inner(cls):
            if (name in self._name_to_query_type_type) and (not overwrite):
                if self._name_to_query_type_type[name] is cls:
                    # redundant registration; do nothing
                    return
                raise Exception(
                    f"The class {self._name_to_query_type_type[name]} is registered to the "
                    f"name {name}. To overwrite, set overwrite=True."
                )
            if cls in self._name_to_query_type_type.values():
                raise Exception(
                    f"The class {cls} is already registered by another name."
                )
            for field in fields(cls):
                if "___" in field.name:
                    raise Exception(
                        "Fields must not contain triple underscores ('___')."
                    )
                    # Why? This would create ambiguity in the server's handling of
                    # search requests. Route signature parameters are named like
                    # "filter___{name}___{field}".
            if cls in self._query_type_to_name:
                raise Exception(
                    f"The type {cls} is already registered to the name "
                    f"{self._query_type_to_name[cls]} and cannot also be "
                    f"registered to {name}."
                )
            self._name_to_query_type_type[name] = cls
            self._query_type_to_name[cls] = name
            return cls

        return inner


# Make a global registry.
query_registry = QueryRegistry()
register = query_registry.register
"""Register a new type of query."""


class QueryTranslationRegistry:
    def __init__(self):
        self._lookup = {}
        self._lazy = {}

    def register(self, class_, translator):
        self._lookup[class_] = translator
        return translator

    def register_lazy(self, toplevel, register):
        """
        Register a registration function which will be called if the
        *toplevel* module (e.g. 'pandas') is ever loaded.
        """
        self._lazy[toplevel] = register

    def dispatch(self, class_):
        """Return the function implementation for the given ``class_``"""
        # Fast path with direct lookup on cls
        lk = self._lookup
        try:
            impl = lk[class_]
        except KeyError:
            pass
        else:
            return impl
        # Is a lazy registration function present?
        toplevel, _, _ = class_.__module__.partition(".")
        try:
            register = self._lazy.pop(toplevel)
        except KeyError:
            pass
        else:
            register()
            return self.dispatch(class_)  # recurse
        # Walk the MRO and cache the lookup result
        for base in inspect.getmro(class_)[1:]:
            if base in lk:
                lk[class_] = lk[base]
                return lk[base]
        raise UnsupportedQueryType(class_.__name__)

    def __call__(self, arg, *args, **kwargs):
        """
        Call the corresponding method based on type of argument.
        """
        meth = self.dispatch(type(arg))
        return meth(arg, *args, **kwargs)



"""
Iterables for KeysView, ValuesView, ItemsView that are sliceable.
"""


class IterViewBase:
    __slots__ = ("_get_length",)

    def __init__(self, get_length):
        self._get_length = get_length

    def __repr__(self):
        return f"<{type(self).__name__}>"

    # Convenience aliases

    def first(self):
        return self[0]

    def last(self):
        return self[-1]

    def head(self, n=5):
        return self[:n]

    def tail(self, n=5):
        return list(reversed(self[-1 : -(n + 1) : -1]))  # noqa: E203

    def __init_subclass__(cls, *args, **kwargs):
        cls.first.__doc__ = f"Get the first {cls._name}."
        cls.last.__doc__ = f"Get the last {cls._name}."
        cls.head.__doc__ = f"Get the first N {cls._name}s."
        cls.tail.__doc__ = f"Get the last N {cls._name}s."

    def __len__(self):
        return self._get_length()


class KeysView(IterViewBase):
    """
    A sliceable, iterable view of keys.
    """

    __slots__ = ("_keys_slice",)
    _name = "key"

    def __init__(self, get_length, keys_slice):
        self._keys_slice = keys_slice
        super().__init__(get_length)

    def __getitem__(self, index_or_slice):
        if isinstance(index_or_slice, int):
            if index_or_slice < 0:
                index_or_slice = -1 - index_or_slice
                direction = -1
            else:
                direction = 1
            keys = list(self._keys_slice(index_or_slice, 1 + index_or_slice, direction))
            try:
                (key,) = keys
            except ValueError:
                raise IndexError("Index out of range")
            return key
        elif isinstance(index_or_slice, slice):
            start, stop, direction = slice_to_interval(index_or_slice)
            return list(self._keys_slice(start, stop, direction))
        else:
            raise TypeError(
                f"{index_or_slice} must be an int or slice, not {type(index_or_slice)}"
            )

    def __iter__(self):
        yield from self._keys_slice(0, None, 1)


class ItemsView(IterViewBase):
    """
    A sliceable, iterable view of (key, value) pairs.
    """

    __slots__ = ("_items_slice",)
    _name = "item"

    def __init__(self, get_length, items_slice):
        self._items_slice = items_slice
        super().__init__(get_length)

    def __getitem__(self, index_or_slice):
        if isinstance(index_or_slice, int):
            if index_or_slice < 0:
                index_or_slice = -1 - index_or_slice
                direction = -1
            else:
                direction = 1
            items = list(
                self._items_slice(index_or_slice, 1 + index_or_slice, direction)
            )
            try:
                (item,) = items
            except ValueError:
                raise IndexError("Index out of range")
            return item
        elif isinstance(index_or_slice, slice):
            start, stop, direction = slice_to_interval(index_or_slice)
            return list(self._items_slice(start, stop, direction))
        else:
            raise TypeError(
                f"{index_or_slice} must be an int or slice, not {type(index_or_slice)}"
            )

    def __iter__(self):
        yield from self._items_slice(0, None, 1)


class ValuesView(IterViewBase):
    """
    A sliceable, iterable view of values.
    """

    __slots__ = ("_items_slice",)
    _name = "value"

    def __init__(self, get_length, items_slice):
        self._items_slice = items_slice
        super().__init__(get_length)

    def __getitem__(self, index_or_slice):
        if isinstance(index_or_slice, int):
            if index_or_slice < 0:
                index_or_slice = -1 - index_or_slice
                direction = -1
            else:
                direction = 1
            items = list(
                self._items_slice(index_or_slice, 1 + index_or_slice, direction)
            )
            try:
                (item,) = items
            except ValueError:
                raise IndexError("Index out of range")
            _key, value = item
            return value
        elif isinstance(index_or_slice, slice):
            start, stop, direction = slice_to_interval(index_or_slice)
            return [value for _key, value in self._items_slice(start, stop, direction)]
        else:
            raise TypeError(
                f"{index_or_slice} must be an int or slice, not {type(index_or_slice)}"
            )

    def __iter__(self):
        for key, value in self._items_slice(0, None, 1):
            yield value


def slice_to_interval(slice_):
    """
    Convert slice object to (start, stop, direction).
    """
    step = slice_.step if slice_.step is not None else 1
    if step == 1:
        start = slice_.start if slice_.start is not None else 0
        if start < 0:
            raise ValueError(
                "Tree sequence slices with start < 0 must have step=-1. "
                f"Use for example [{slice_.start}:{slice_.stop}:-1]"
                "(This is a limitation of slicing on Tree sequences "
                "that does not apply to Python sequences in general.)"
            )
        if (slice_.stop is not None) and (slice_.stop < start):
            raise ValueError(
                "Tree sequence slices with step=1 must have stop >= start. "
                "(This is a limitation of slicing on Tree sequences "
                "that does not apply to Python sequences in general.)"
            )
        start_ = start
        stop_ = slice_.stop
        direction = 1
    elif step == -1:
        start = slice_.start if slice_.start is not None else -1
        if start >= 0:
            raise ValueError(
                "Tree sequence slices with start >= 0 must have step=1. "
                "(This is a limitation of slicing on Tree sequences "
                "that does not apply to Python sequences in general.)"
            )
        if slice_.stop is not None:
            if slice_.stop > start:
                raise ValueError(
                    "Tree sequence slices with step=-1 must have stop <= start."
                )
            stop_ = -(slice_.stop + 1)
        else:
            stop_ = slice_.stop
        start_ = -(start + 1)
        direction = -1
    else:
        raise ValueError(
            "Only step of 1 or -1 is supported in a Tree sequence slice. "
            f"Step {slice_.step} is disallowed."
        )
    assert start_ >= 0
    assert (stop_ is None) or (stop_ >= start_)
    return start_, stop_, direction



SCOPES = {
    "read:metadata": {"description": "Read metadata."},
    "read:data": {"description": "Read data."},
    "write:metadata": {"description": "Write metadata."},
    "write:data": {"description": "Write data."},
    "create": {"description": "Add a node."},
    "metrics": {"description": "Access (Prometheus) metrics."},
    "apikeys": {
        "description": "Create and revoke API keys as the currently-authenticated user or service."
    },
    "admin:apikeys": {
        "description": "Create and revoke API keys on behalf of any user or service."
    },
    "read:principals": {
        "description": "Read list of all users and services and their attributes."
    },
    "write:principals": {
        "description": "Edit list of all users and services and their attributes."
    },
}



from typing import List

from sqlalchemy import (
    JSON,
    Boolean,
    Column,
    DateTime,
    Enum,
    ForeignKey,
    Index,
    Integer,
    Unicode,
    event,
    text,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.schema import UniqueConstraint
from sqlalchemy.sql import func

from ..server.schemas import Management
from ..structures.core import StructureFamily
from .base import Base

# Use JSON with SQLite and JSONB with PostgreSQL.
JSONVariant = JSON().with_variant(JSONB(), "postgresql")


class Timestamped:
    """
    Mixin for providing timestamps of creation and update time.

    These are not used by application code, but they may be useful for
    forensics.
    """

    time_created = Column(DateTime(timezone=False), server_default=func.now())
    time_updated = Column(
        DateTime(timezone=False), onupdate=func.now(), server_default=func.now()
    )

    def __repr__(self):
        return (
            f"{type(self).__name__}("
            + ", ".join(
                f"{key}={value!r}"
                for key, value in self.__dict__.items()
                if not key.startswith("_")
            )
            + ")"
        )


class Node(Timestamped, Base):
    """
    This describes a single Node and sometimes inlines descriptions of all its children.
    """

    __tablename__ = "nodes"
    __mapper_args__ = {"eager_defaults": True}

    # This id is internal, never exposed to the client.
    id = Column(Integer, primary_key=True, index=True, autoincrement=True)

    key = Column(Unicode(1023), nullable=False)
    ancestors = Column(JSONVariant, nullable=True)
    structure_family = Column(Enum(StructureFamily), nullable=False)
    metadata_ = Column("metadata", JSONVariant, nullable=False)
    specs = Column(JSONVariant, nullable=False)

    data_sources = relationship(
        "DataSource",
        backref="node",
        cascade="save-update",
        lazy="selectin",
        passive_deletes=True,
    )
    revisions = relationship(
        "Revision",
        backref="revisions",
        passive_deletes=True,
    )

    __table_args__ = (
        UniqueConstraint("key", "ancestors", name="key_ancestors_unique_constraint"),
        # This index supports comparison operations (==, <, ...).
        # For key-existence operations we will need a GIN index additionally.
        Index(
            "top_level_metadata",
            "ancestors",
            # include the keys of the default sorting ('time_created', 'id'),
            # used to avoid creating a temp sort index
            "time_created",
            "id",
            "metadata",
            postgresql_using="gin",
        )
        # This is used by ORDER BY with the default sorting.
        # Index("ancestors_time_created", "ancestors", "time_created"),
    )


class DataSourceAssetAssociation(Base):
    """
    This describes which Assets are used by which DataSources, and how.

    The 'parameter' describes which argument to pass the asset into when
    constructing the Adapter. If 'parameter' is NULL then the asset is an
    indirect dependency, such as a HDF5 data file backing an HDF5 'master'
    file.

    If 'num' is NULL, the asset is passed as a scalar value, and there must be
    only one for the given 'parameter'. If 'num' is not NULL, all the assets
    sharing the same 'parameter' (there may be one or more) will be passed in
    as a list, ordered in ascending order of 'num'.
    """

    __tablename__ = "data_source_asset_association"

    data_source_id: Mapped[int] = mapped_column(
        ForeignKey("data_sources.id", ondelete="CASCADE"),
        primary_key=True,
    )
    asset_id: Mapped[int] = mapped_column(
        ForeignKey("assets.id", ondelete="CASCADE"),
        primary_key=True,
    )
    parameter = Column(Unicode(255), nullable=True)
    num = Column(Integer, nullable=True)

    data_source: Mapped["DataSource"] = relationship(
        back_populates="asset_associations"
    )
    asset: Mapped["Asset"] = relationship(
        back_populates="data_source_associations", lazy="selectin"
    )

    __table_args__ = (
        UniqueConstraint(
            "data_source_id",
            "parameter",
            "num",
            name="parameter_num_unique_constraint",
        ),
        # Below, in unique_parameter_num_null_check, additional constraints
        # are applied, via triggers.
    )


@event.listens_for(DataSourceAssetAssociation.__table__, "after_create")
def unique_parameter_num_null_check(target, connection, **kw):
    # This creates a pair of triggers on the data_source_asset_association
    # table. (There are a total of four defined below, two for the SQLite
    # branch and two for the PostgreSQL branch.) Each pair include one trigger
    # that runs when NEW.num IS NULL and one trigger than runs when
    # NEW.num IS NOT NULL. Thus, for a given insert, only one of these
    # triggers is run.
    if connection.engine.dialect.name == "sqlite":
        connection.execute(
            text(
                """
CREATE TRIGGER cannot_insert_num_null_if_num_exists
BEFORE INSERT ON data_source_asset_association
WHEN NEW.num IS NULL
BEGIN
    SELECT RAISE(ABORT, 'Can only insert num=NULL if no other row exists for the same parameter')
    WHERE EXISTS
    (
        SELECT 1
        FROM data_source_asset_association
        WHERE parameter = NEW.parameter
        AND data_source_id = NEW.data_source_id
    );
END"""
            )
        )
        connection.execute(
            text(
                """
CREATE TRIGGER cannot_insert_num_int_if_num_null_exists
BEFORE INSERT ON data_source_asset_association
WHEN NEW.num IS NOT NULL
BEGIN
    SELECT RAISE(ABORT, 'Can only insert INTEGER num if no NULL row exists for the same parameter')
    WHERE EXISTS
    (
        SELECT 1
        FROM data_source_asset_association
        WHERE parameter = NEW.parameter
        AND num IS NULL
        AND data_source_id = NEW.data_source_id
    );
END"""
            )
        )
    elif connection.engine.dialect.name == "postgresql":
        connection.execute(
            text(
                """
CREATE OR REPLACE FUNCTION raise_if_parameter_exists()
RETURNS TRIGGER AS $$
BEGIN
    IF EXISTS (
        SELECT 1
        FROM data_source_asset_association
        WHERE parameter = NEW.parameter
        AND data_source_id = NEW.data_source_id
    ) THEN
        RAISE EXCEPTION 'Can only insert num=NULL if no other row exists for the same parameter';
    END IF;
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;"""
            )
        )
        connection.execute(
            text(
                """
CREATE TRIGGER cannot_insert_num_null_if_num_exists
BEFORE INSERT ON data_source_asset_association
FOR EACH ROW
WHEN (NEW.num IS NULL)
EXECUTE FUNCTION raise_if_parameter_exists();"""
            )
        )
        connection.execute(
            text(
                """
CREATE OR REPLACE FUNCTION raise_if_null_parameter_exists()
RETURNS TRIGGER AS $$
BEGIN
    IF EXISTS (
        SELECT 1
        FROM data_source_asset_association
        WHERE parameter = NEW.parameter
        AND data_source_id = NEW.data_source_id
        AND num IS NULL
    ) THEN
        RAISE EXCEPTION 'Can only insert INTEGER num if no NULL row exists for the same parameter';
    END IF;
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;"""
            )
        )
        connection.execute(
            text(
                """
CREATE TRIGGER cannot_insert_num_int_if_num_null_exists
BEFORE INSERT ON data_source_asset_association
FOR EACH ROW
WHEN (NEW.num IS NOT NULL)
EXECUTE FUNCTION raise_if_null_parameter_exists();"""
            )
        )


@event.listens_for(DataSourceAssetAssociation.__table__, "after_create")
def create_index_metadata_tsvector_search(target, connection, **kw):
    # This creates a ts_vector based metadata search index for fulltext.
    # Postgres only feature
    if connection.engine.dialect.name == "postgresql":
        connection.execute(
            text(
                """
                CREATE INDEX metadata_tsvector_search
                ON nodes
                USING gin (jsonb_to_tsvector('simple', metadata, '["string"]'))
                """
            )
        )


class DataSource(Timestamped, Base):
    """
    The describes how to open one or more file/blobs to extract data for a Node.

    The mimetype can be used to look up an appropriate Adapter.
    The Adapter will accept the data_uri (which may be a directory in this case)
    and optional parameters.

    The parameters are used to select the data of interest for this DataSource.
    Then, within that, Tiled may use the standard Adapter API to subselect the data
    of interest for a given request.
    """

    __tablename__ = "data_sources"
    __mapper_args__ = {"eager_defaults": True}

    id: Mapped[int] = mapped_column(primary_key=True, index=True, autoincrement=True)
    node_id = Column(
        Integer, ForeignKey("nodes.id", ondelete="CASCADE"), nullable=False
    )
    structure_id = Column(
        Unicode(32), ForeignKey("structures.id", ondelete="CASCADE"), nullable=True
    )
    mimetype = Column(Unicode(255), nullable=False)  # max length given by RFC 4288
    # These are additional parameters passed to the Adapter to guide
    # it to access and arrange the data in the file correctly.
    parameters = Column(JSONVariant, nullable=True)
    # This relates to the mutability of the data.
    management = Column(Enum(Management), nullable=False)

    # many-to-one relationship to Structure
    structure: Mapped["Structure"] = relationship(
        "Structure",
        lazy="selectin",
        passive_deletes=True,
    )

    # many-to-many relationship to Asset, bypassing the `Association` class
    assets: Mapped[List["Asset"]] = relationship(
        secondary="data_source_asset_association",
        back_populates="data_sources",
        cascade="all, delete",
        lazy="selectin",
        viewonly=True,
    )
    # association between Asset -> Association -> DataSource
    asset_associations: Mapped[List["DataSourceAssetAssociation"]] = relationship(
        back_populates="data_source",
        lazy="selectin",
        order_by=[DataSourceAssetAssociation.parameter, DataSourceAssetAssociation.num],
    )


class Structure(Base):
    """
    The describes the structure of a DataSource.

    The id is the HEX digest of the MD5 hash of the canonical representation
    of the JSON structure, as specified by RFC 8785.

    https://datatracker.ietf.org/doc/html/rfc8785

    """

    __tablename__ = "structures"

    id: str = Column(Unicode(32), primary_key=True, unique=True)
    structure = Column(JSONVariant, nullable=False)


class Asset(Timestamped, Base):
    """
    This tracks individual files/blobs.
    """

    __tablename__ = "assets"
    __mapper_args__ = {"eager_defaults": True}

    id: Mapped[int] = mapped_column(primary_key=True, index=True, autoincrement=True)

    data_uri = Column(Unicode(1023), index=True, unique=True)
    is_directory = Column(Boolean, nullable=False)
    hash_type = Column(Unicode(63), nullable=True)
    hash_content = Column(Unicode(1023), nullable=True)
    size = Column(Integer, nullable=True)

    # # many-to-many relationship to Asset, bypassing the `Association` class
    data_sources: Mapped[List["DataSource"]] = relationship(
        secondary="data_source_asset_association",
        back_populates="assets",
        viewonly=True,
    )
    # association between DataSource -> Association -> Asset
    data_source_associations: Mapped[List["DataSourceAssetAssociation"]] = relationship(
        back_populates="asset",
    )


class Revision(Timestamped, Base):
    """
    This tracks history of metadata and specs, supporting 'undo' functionality.
    """

    __tablename__ = "revisions"
    __mapper_args__ = {"eager_defaults": True}

    # This id is internal, never exposed to the client.
    id = Column(Integer, primary_key=True, index=True, autoincrement=True)

    node_id = Column(
        Integer, ForeignKey("nodes.id", ondelete="CASCADE"), nullable=False
    )
    revision_number = Column(Integer, nullable=False)

    metadata_ = Column("metadata", JSONVariant, nullable=False)
    specs = Column(JSONVariant, nullable=False)

    __table_args__ = (
        UniqueConstraint(
            "node_id",
            "revision_number",
            name="node_id_revision_number_unique_constraint",
        ),
    )
